import importlib
import inspect
import logging
import os
import shutil
import sys
import time
import traceback
from pathlib import Path

from cognite.client import CogniteClient
from prometheus_client import Gauge
from pydantic import BaseModel

from cognite.neat.workflows import BaseWorkflow
from cognite.neat.workflows.base import WorkflowDefinition
from cognite.neat.workflows.model import FlowMessage, InstanceStartMethod, WorkflowState
from cognite.neat.workflows.steps_registry import StepsRegistry
from cognite.neat.workflows.tasks import WorkflowTaskBuilder

live_workflow_instances = Gauge("neat_workflow_live_instances", "Count of live workflow instances", ["itype"])


class WorkflowStartStatus(BaseModel):
    workflow_instance: BaseWorkflow = None
    is_success: bool = True
    status_text: str = None

    class Config:
        arbitrary_types_allowed = True


class WorkflowManager:
    """Workflow manager is responsible for loading, saving and managing workflows
    client: CogniteClient
    registry_storage_type: str = "file"
    workflows_storage_path: Path = Path("workflows")
    rules_storage_path: Path = Path("rules")
    data_set_id: int = None,
    """

    def __init__(
        self,
        client: CogniteClient = None,
        registry_storage_type: str = "file",
        workflows_storage_path: Path = None,
        rules_storage_path: Path = None,
        data_store_path: Path = None,
        data_set_id: int = None,
    ):
        self.client = client
        self.data_set_id = data_set_id
        self.data_store_path = data_store_path
        self.workflow_registry: dict[str, BaseWorkflow] = {}
        self.ephemeral_instance_registry: dict[str, BaseWorkflow] = {}
        self.workflows_storage_type = registry_storage_type
        # todo use pathlib
        self.workflows_storage_path = workflows_storage_path if workflows_storage_path else Path("workflows")
        self.rules_storage_path = rules_storage_path if rules_storage_path else Path("rules")
        self.task_builder = WorkflowTaskBuilder(client, self)
        self.steps_registry = StepsRegistry(self.data_store_path)
        self.steps_registry.load_step_classes()

    def update_cdf_client(self, client: CogniteClient):
        self.client = client
        self.task_builder = WorkflowTaskBuilder(client, self)
        self.workflow_registry = {}
        self.load_workflows_from_storage()

    def get_list_of_workflows(self):
        return list(self.workflow_registry.keys())

    def get_steps_registry(self):
        return self.steps_registry

    def get_workflow(self, name: str) -> BaseWorkflow | None:
        try:
            return self.workflow_registry[name]
        except KeyError:
            return None

    def start_workflow(self, name: str, sync=False, **kwargs):
        workflow = self.get_workflow(name)
        workflow.start(sync, kwargs=kwargs)
        return workflow

    def delete_workflow(self, name: str):
        del self.workflow_registry[name]
        full_path = self.workflows_storage_path / name
        shutil.rmtree(full_path)
        # TODO: check if more garbage collection is needed here.

    def update_workflow(self, name: str, workflow: WorkflowDefinition):
        self.workflow_registry[name].workflow_steps = workflow.steps
        self.workflow_registry[name].configs = workflow.configs
        self.workflow_registry[name].workflow_system_components = workflow.system_components

    def save_workflow_to_storage(self, name: str, custom_implementation_module: str = None):
        """Save workflow from memory to storage"""
        if self.workflows_storage_type == "file":
            full_path = self.workflows_storage_path / name / "workflow.yaml"
            full_path.parent.mkdir(parents=True, exist_ok=True)
            wf = self.workflow_registry[name]
            with open(full_path, "w") as f:
                f.write(
                    wf.serialize_workflow(
                        output_format="yaml", custom_implementation_module=custom_implementation_module
                    )
                )

    def create_new_workflow(self, name: str, description=None, mode: str = "manifest") -> WorkflowDefinition:
        """Create new workflow in memory"""
        workflow_definition = WorkflowDefinition(
            name=name, description=description, steps=[], system_components=[], configs=[]
        )
        self.register_workflow(BaseWorkflow, name, workflow_definition)
        self.save_workflow_to_storage(name)
        return workflow_definition

    def load_workflows_from_storage(self, workflows_storage_path: str | Path = None):
        """Loads workflows from disk/storage into memory, initializes and register them in the workflow registry"""

        # set workflow storage path
        workflows_storage_path = Path(workflows_storage_path) if workflows_storage_path else self.workflows_storage_path

        # set system path to be used when importing individual workflows as python modules
        # via importlib.import_module(...)
        sys.path.append(str(workflows_storage_path))

        for workflow_name in os.listdir(workflows_storage_path):
            workflow_path = workflows_storage_path / workflow_name
            # THIS IS WHERE WE LOAD THE WORKFLOW MODULES
            if workflow_path.is_dir():
                logging.info(f"Loading workflow {workflow_name} from {workflow_path}")
                try:
                    workflow_definition_path = workflow_path / "workflow.yaml"
                    if os.path.exists(workflow_definition_path):
                        with open(workflow_definition_path, "r") as workflow_definition_file:
                            workflow_definition: WorkflowDefinition = BaseWorkflow.deserialize_definition(
                                workflow_definition_file.read(), output_format="yaml"
                            )
                    else:
                        logging.info(f"Definition file {workflow_definition_path} not found, skipping")
                        continue

                    # This is convinience feature when user wants to share self contained workflow
                    # folder as single zip file.
                    if (workflow_path / "rules").exists():
                        logging.info(f"Copying rules from {workflow_path / 'rules'} to {self.rules_storage_path} ")
                        for rule_file in os.listdir(workflow_path / "rules"):
                            shutil.copy(workflow_path / "rules" / rule_file, self.rules_storage_path)

                    # Comment: All our workflows implementation_module is None
                    # what is this meant for ?, just to have different name?
                    if workflow_definition.implementation_module:
                        workflow_name = workflow_definition.implementation_module
                        logging.info(f"Loading CUSTOM workflow module {workflow_name}")
                    else:
                        logging.info(f"Loading workflow module {workflow_name}")

                    full_module_name = f"{workflow_name}.workflow"
                    load_user_defined_workflow = False
                    if full_module_name in sys.modules:
                        logging.info(f"Reloading existing workflow module {workflow_name}")
                        module = importlib.reload(sys.modules[full_module_name])
                        load_user_defined_workflow = True
                    else:
                        try:
                            module = importlib.import_module(full_module_name)
                            load_user_defined_workflow = True
                            logging.info(f"Workflow implementation class for {workflow_name} loaded successfully")
                        except ModuleNotFoundError:
                            pass

                    self.steps_registry.load_custom_step_classes(workflow_path, workflow_name)
                    # Dynamically load workflow classes which contain "NeatWorkflow" in their name
                    # from workflow.py module in the workflow directory and
                    # Instantiate them using the workflow definition loaded
                    # from workflow.yaml file
                    # WARNING: This will be deprecated in the future.
                    if load_user_defined_workflow:
                        for name, obj in inspect.getmembers(module):
                            if "NeatWorkflow" in name and inspect.isclass(obj):
                                logging.info(
                                    (
                                        f"Found class {name} in module {workflow_name},"
                                        f" registering it as '{workflow_name}' in the workflow registry"
                                    )
                                )
                                self.register_workflow(obj, workflow_name, workflow_definition)
                                return
                    else:
                        self.register_workflow(BaseWorkflow, workflow_name, workflow_definition)

                except Exception as e:
                    trace = traceback.format_exc()
                    logging.error(f"Error loading workflow {workflow_name}: error: {e} trace : {trace}")

    def register_workflow(self, obj, workflow_name, workflow_definition):
        """Register workflow in the workflow registry

        Parameters
        ----------
        obj : BaseWorkflow
            Class of the workflow to be registered
        workflow_name : _type_
            Name of the workflow to be registered
        workflow_definition : _type_
            Definition of the workflow to be registered originating from workflow.yaml
        """
        self.workflow_registry[workflow_name] = obj(workflow_name, self.client, steps_registry=self.steps_registry)
        self.workflow_registry[workflow_name].set_definition(workflow_definition)
        # Comment: Not entirely sure what is task_builder meant for
        # as at first glance it looks like circular import ???
        self.workflow_registry[workflow_name].set_task_builder(self.task_builder)
        self.workflow_registry[workflow_name].set_default_dataset_id(self.data_set_id)
        self.workflow_registry[workflow_name].set_storage_path("transformation_rules", self.rules_storage_path)
        self.workflow_registry[workflow_name].set_storage_path("data_store", self.data_store_path)

    def create_workflow_instance(self, template_name: str, add_to_registry: bool = True) -> BaseWorkflow:
        new_instance = self.workflow_registry[template_name].copy()
        if add_to_registry:
            self.ephemeral_instance_registry[new_instance.instance_id] = new_instance
        live_workflow_instances.labels(itype="ephemeral").set(len(self.ephemeral_instance_registry))
        return new_instance

    def get_workflow_instance(self, instance_id: str) -> BaseWorkflow:
        return self.ephemeral_instance_registry[instance_id]

    def delete_workflow_instance(self, instance_id: str):
        del self.ephemeral_instance_registry[instance_id]
        live_workflow_instances.labels(itype="ephemeral").set(len(self.ephemeral_instance_registry))
        return

    def start_workflow_instance(
        self, workflow_name: str, step_id: str = "", flow_msg: FlowMessage = None, sync: bool = None
    ) -> WorkflowStartStatus:
        workflow = self.get_workflow(workflow_name)

        trigger_step = workflow.get_trigger_step(step_id)
        if not trigger_step.trigger:
            logging.info(f"Step {step_id} is not a trigger step")
            return WorkflowStartStatus(
                workflow_instance=None, is_success=False, status_text="Step is not a trigger step"
            )
        if sync is None:
            sync = trigger_step.params.get("sync", "true").lower() == "true"

        max_wait_time = int(trigger_step.params.get("max_wait_time", "30"))
        instance_start_method = trigger_step.params.get(
            "workflow_start_method", InstanceStartMethod.PERSISTENT_INSTANCE_BLOCKING
        )

        logging.info(
            f"""----- Starting workflow {workflow_name} , step_id = {step_id} , sync = {sync},
              max_wait_time = {max_wait_time}, instance_start_method = {instance_start_method} -----"""
        )

        if instance_start_method == InstanceStartMethod.PERSISTENT_INSTANCE_BLOCKING:
            live_workflow_instances.labels(itype="persistent").set(len(self.workflow_registry))
            # wait until workflow transition to RUNNING state and then start , set max wait time to 30 seconds
            start_time = time.perf_counter()
            # wait until workflow transition to RUNNING state and then start , set max wait time to 30 seconds.
            # The operation is executed in callers thread
            while workflow.state == WorkflowState.RUNNING:
                logging.info("Existing workflow instance already running , waiting for RUNNING state")
                elapsed_time = time.perf_counter() - start_time
                if elapsed_time > max_wait_time:
                    logging.info(
                        f"Workflow {workflow_name} wait time exceeded . "
                        f"elapsed time = {elapsed_time}, max wait time = {max_wait_time}"
                    )
                    return WorkflowStartStatus(
                        workflow_instance=None,
                        is_success=False,
                        status_text="Workflow instance already running.Wait time exceeded",
                    )
                time.sleep(0.5)
            workflow.start(sync=sync, flow_message=flow_msg, start_step_id=step_id)
            return WorkflowStartStatus(workflow_instance=workflow, is_success=True, status_text="")

        elif instance_start_method == InstanceStartMethod.PERSISTENT_INSTANCE_NON_BLOCKING:
            live_workflow_instances.labels(itype="persistent").set(len(self.workflow_registry))
            # start workflow if not already running, skip if already running
            if workflow.state == WorkflowState.RUNNING:
                return WorkflowStartStatus(
                    workflow_instance=None, is_success=False, status_text="Workflow instance already running"
                )

            workflow.start(sync=sync, flow_message=flow_msg, start_step_id=step_id)
            return WorkflowStartStatus(workflow_instance=workflow, is_success=True, status_text="")

        elif instance_start_method == InstanceStartMethod.EPHEMERAL_INSTANCE:
            # start new workflow instance in new thread
            workflow = self.create_workflow_instance(workflow_name, add_to_registry=True)
            workflow.start(sync=sync, delete_after_completion=True, flow_message=flow_msg, start_step_id=step_id)
            if sync:
                self.delete_workflow_instance(workflow.instance_id)
            return WorkflowStartStatus(workflow_instance=workflow, is_success=True, status_text="")

        return WorkflowStartStatus(
            workflow_instance=None, is_success=False, status_text="Unsupported workflow start method"
        )



import json
import logging
import threading
import time

import schedule
from fastapi import Depends, FastAPI, Request

from cognite.neat.workflows.manager import WorkflowManager
from cognite.neat.workflows.model import FlowMessage, StepType, WorkflowState


async def get_body(request: Request):
    return await request.body()


fast_api_depends = Depends(get_body)


class TriggerManager:
    """Triggers are the way to start a workflow. They are defined in the workflow definition file."""

    def __init__(self, workflow_manager: WorkflowManager):
        self.workflow_manager = workflow_manager
        self.is_running = False
        self.thread = None

    def start_http_listeners(self, web_server: FastAPI):
        """Starts a HTTP listener for the workflows

        Parameters
        ----------
        web_server : FastAPI instance
        """
        logging.info("Starting HTTP trigger endpoint")

        @web_server.post("/api/workflow/{workflow_name}/http_trigger/{step_id}")
        def start_workflow(workflow_name: str, step_id: str, request: Request, body: bytes = fast_api_depends):
            logging.info(f"New HTTP trigger request for workflow {workflow_name} step {step_id}")
            headers = dict(request.headers)
            logging.debug(f"Request headers = {headers}")
            json_payload = None
            try:
                json_payload = json.loads(body)
            except Exception as e:
                logging.info(f"Error parsing json body {e}")
            logging.debug(f"Request object = {json_payload}")

            flow_msg = FlowMessage(payload=json_payload, headers=dict(headers))
            start_status = self.workflow_manager.start_workflow_instance(
                workflow_name=workflow_name, step_id=step_id, flow_msg=flow_msg
            )
            if start_status.is_success:
                return start_status.workflow_instance.flow_message

        @web_server.post("/api/workflow/{workflow_name}/resume/{step_id}/{instance_id}")
        def resume_workflow(
            workflow_name: str, step_id: str, instance_id: str, request: Request, body: bytes = fast_api_depends
        ):
            if instance_id != "default":
                workflow = self.workflow_manager.get_workflow_instance(instance_id)
            else:
                workflow = self.workflow_manager.get_workflow(workflow_name)

            json_payload = None
            try:
                json_payload = json.loads(body)
            except ValueError as e:
                logging.info(f"Error parsing json body {e}")
            flow_msg = FlowMessage(payload=json_payload)
            if workflow.state == WorkflowState.RUNNING_WAITING:
                workflow.resume_workflow(flow_message=flow_msg, step_id=step_id)
                return {"result": "Workflow instance resumed"}

            return {"result": "Workflow instance not in RUNNING_WAITING state"}

    def _start_scheduler_main_loop(self):
        """Starts a scheduler main loop for the workflows

        Parameters
        ----------

        """
        logging.info("Starting scheduler main loop")
        self.is_running = True

        def main_loop():
            while self.is_running:
                schedule.run_pending()
                time.sleep(1)
            logging.info("Scheduler main loop stopped")

        self.thread = threading.Thread(target=main_loop)
        self.thread.start()

    def stop_scheduler_main_loop(self):
        """Stops a scheduler main loop for the workflows

        Parameters
        ----------

        """
        logging.info("Stopping scheduler main loop")
        self.is_running = False
        if self.thread:
            self.thread.join()
        logging.info("Scheduler main loop stopped")

    def start_time_schedulers(self):
        """Starts a time scheduler for the workflows

        Parameters
        ----------

        """
        logging.info("Starting time trigger scheduler")

        trigger_steps_count = 0
        for workflow in self.workflow_manager.workflow_registry.values():
            for step in workflow.workflow_steps:
                if step.trigger and step.stype == StepType.TIME_TRIGGER and step.enabled:
                    # params
                    every_str = step.params["interval"]
                    logging.info(f"Starting time trigger scheduler for {workflow.name} {step.id} {every_str}")
                    every_str = every_str.replace("every", "").strip()

                    if "at" in every_str:  # "day at 10:30:00"
                        every = every_str.split("at ")
                        if len(every) == 2:
                            interval_unit = every[0]
                            interval_value = every[1]
                            if "day" in interval_unit:
                                trigger_steps_count += 1
                                schedule.every().day.at(interval_value).do(workflow.start, start_step_id=step.id)
                        else:
                            logging.error(f"Invalid time trigger interval {every_str}")

                    else:  # "5 minutes"
                        every = every_str.split(" ")
                        if len(every) == 2:
                            interval_unit = every[1]
                            interval_value = every[0]
                            trigger_steps_count += 1
                            if "minutes" in interval_unit:  # "day at 10:30:00" , "5 minutes"
                                schedule.every(int(interval_value)).minutes.do(workflow.start, start_step_id=step.id)
                            elif "hours" in interval_unit:
                                schedule.every(int(interval_value)).hours.do(workflow.start, start_step_id=step.id)
                            elif "days" in interval_unit:
                                schedule.every(int(interval_value)).days.do(workflow.start, start_step_id=step.id)
                            elif "seconds" in interval_unit:
                                schedule.every(int(interval_value)).seconds.do(workflow.start, start_step_id=step.id)
                            else:
                                logging.error(f"Invalid time trigger interval {every_str}")
                                trigger_steps_count -= 1
                        else:
                            logging.error(f"Invalid time trigger interval {every_str}")

        if trigger_steps_count > 0 and not self.is_running:
            self._start_scheduler_main_loop()
            logging.info("ALl Time trigger scheduler started")
        else:
            logging.info("No Time trigger scheduler started")

    def reload_all_triggers(self):
        """Reloads all triggers

        Parameters
        ----------
        """
        logging.info("Reloading all triggers")
        schedule.clear()
        self.is_running = False
        time.sleep(1)
        self.start_time_schedulers()



import contextlib
import logging
import time
from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes import AssetFilter
from prometheus_client import Gauge

from cognite.neat import rules
from cognite.neat.graph import extractors
from cognite.neat.graph.loaders.core.labels import upload_labels
from cognite.neat.graph.loaders.core.rdf_to_assets import (
    NeatMetadataKeys,
    categorize_assets,
    rdf2assets,
    remove_non_existing_labels,
    unique_asset_labels,
    upload_assets,
)
from cognite.neat.graph.loaders.core.rdf_to_relationships import (
    categorize_relationships,
    rdf2relationships,
    upload_relationships,
)
from cognite.neat.graph.extractors import NeatGraphStore
from cognite.neat.rules.exporter.rules2triples import get_instances_as_triples
from cognite.neat.rules.models import TransformationRules
from cognite.neat.graph.loaders.validator import validate_asset_hierarchy
from cognite.neat.workflows import utils
from cognite.neat.workflows.base import BaseWorkflow
from cognite.neat.workflows.model import FlowMessage
from cognite.neat.workflows.cdf_store import CdfStore

with contextlib.suppress(ValueError):
    prom_cdf_resource_stats = Gauge(
        "neat_sheet2cdf_cdf_resource_stats",
        "CDF resource stats before and after running sheet2cdf workflow",
        ["resource_type", "state"],
    )
with contextlib.suppress(ValueError):
    prom_data_issues_stats = Gauge("neat_sheet2cdf_wf_data_issues", "Data validation issues", ["type"])


class Sheet2CDFBaseWorkflow(BaseWorkflow):
    def __init__(self, name: str, client: CogniteClient):
        super().__init__(name, client, [])
        self.dataset_id: int = 0
        self.current_step: str = None
        self.source_graph: NeatGraphStore = None
        self.solution_graph: NeatGraphStore = None
        self.raw_tables = None
        self.transformation_rules: TransformationRules = None
        self.stop_on_error = False
        self.triples = []
        self.instance_ids = set()
        self.count_create_assets = 0
        self.meta_keys: NeatMetadataKeys | None = None

    def step_load_transformation_rules(self, flow_msg: FlowMessage = None):
        # Load rules from file or remote location
        cdf_store = CdfStore(self.cdf_client, self.dataset_id, rules_storage_path=self.rules_storage_path)

        rules_file = self.get_config_item("rules.file").value
        rules_file_path = Path(self.rules_storage_path, rules_file)
        version = self.get_config_item("rules.version").value

        if rules_file_path.exists() and not version:
            logging.info(f"Loading rules from {rules_file_path}")
        elif rules_file_path.exists() and version:
            hash = utils.get_file_hash(rules_file_path)
            if hash != version:
                cdf_store.load_rules_file_from_cdf(rules_file, version)
        else:
            cdf_store.load_rules_file_from_cdf(self.cdf_client, rules_file, version)

        self.transformation_rules, self.errors, self.warnings = rules.parse_rules_from_excel_file(
            rules_file_path, return_report=True
        )

        output_text = f"Loaded {len(self.transformation_rules.properties)} rules from {rules_file_path.name!r}."
        logging.info(output_text)
        logging.info(f"Loaded prefixes {str(self.transformation_rules.prefixes)} rules")

        self.dataset_id = self.transformation_rules.metadata.data_set_id
        return FlowMessage(output_text=output_text)

    def step_configuring_stores(self, flow_msg: FlowMessage = None, clean_start: bool = True):
        self.source_graph = extractors.NeatGraphStore(
            prefixes=self.transformation_rules.prefixes, namespace=self.transformation_rules.metadata.namespace
        )
        self.source_graph.init_graph(base_prefix=self.transformation_rules.metadata.prefix)

        # this is fix to be able to display the graph in the UI
        # alex is working on a better solution
        self.solution_graph = self.source_graph

        return FlowMessage(output_text="Configured in-memory graph store")

    def step_parse_instances(self, flow_msg: FlowMessage = None):
        # TODO: Need to provide info both as metric and as report about
        # total number of rows in the sheet that have been processed by the workflow
        # and report back reasons why

        self.triples = get_instances_as_triples(self.transformation_rules)
        self.instance_ids = {triple[0] for triple in self.triples}

        output_text = f"Loaded {len(self.instance_ids)} instances out of"
        # Todo: This is no longer exposed in the rules package. Need to extend the load methods to return a rapport.
        # output_text += f" {len(self.raw_tables['Instances'])} rows in Instances sheet"

        logging.info(output_text)
        return FlowMessage(output_text=output_text)

    def step_load_instances_to_source_graph(self, flow_msg: FlowMessage = None):
        # Load parsed instances to source graph

        try:
            for triple in self.triples:
                self.source_graph.graph.add(triple)
        except Exception as e:
            logging.error("Not able to load instances to source graph")
            raise e

        output_text = f"Loaded {len(self.triples)} statements defining"
        output_text += f" {len(self.instance_ids)} instances"

        logging.info(output_text)
        return FlowMessage(output_text=output_text)

    def step_create_cdf_labels(self, flow_msg: FlowMessage = None):
        logging.info("Creating CDF labels")
        upload_labels(self.cdf_client, self.transformation_rules, extra_labels=["non-historic", "historic"])

    def step_prepare_cdf_assets(self, flow_msg: FlowMessage):
        # export graph into CDF
        # TODO : decide on error handling and retry logic\
        self.meta_keys = NeatMetadataKeys.load(
            self.get_config_group_values_by_name("cdf.asset.metadata.", remove_group_prefix=True)
        )

        rdf_assets = rdf2assets(
            self.solution_graph,
            self.transformation_rules,
            stop_on_exception=self.stop_on_error,
            meta_keys=self.meta_keys,
        )

        if not self.cdf_client:
            logging.info("Dry run, no CDF client available")
            return

        # Label Validation
        labels_before = unique_asset_labels(rdf_assets.values())
        logging.info(f"Assets have {len(labels_before)} unique labels: {', '.join(sorted(labels_before))}")

        rdf_assets = remove_non_existing_labels(self.cdf_client, rdf_assets)

        labels_after = unique_asset_labels(rdf_assets.values())
        removed_labels = labels_before - labels_after
        logging.info(
            f"Removed {len(removed_labels)} labels as these do not exists in CDF. "
            f"Removed labels: {', '.join(sorted(removed_labels))}"
        )
        ######################

        # UPDATE: 2023-04-05 - correct aggregation of assets in CDF for specific dataset
        total_assets_before = self.cdf_client.assets.aggregate(
            filter=AssetFilter(data_set_ids=[{"id": self.dataset_id}])
        )[0]["count"]

        prom_cdf_resource_stats.labels(resource_type="asset", state="count_before_neat_update").set(total_assets_before)
        logging.info(f"Total count of assets in CDF before upload: { total_assets_before }")

        orphan_assets, circular_assets = validate_asset_hierarchy(rdf_assets)

        prom_data_issues_stats.labels(type="circular_assets").set(len(circular_assets))
        prom_data_issues_stats.labels(type="orphan_assets").set(len(orphan_assets))

        if orphan_assets:
            logging.error(f"Found orphaned assets: {', '.join(orphan_assets)}")

            orphanage_asset_external_id = (
                f"{self.transformation_rules.metadata.externalIdPrefix}orphanage"
                if self.transformation_rules.metadata.externalIdPrefix
                else "orphanage"
            )

            # Kill the process if you dont have orphanage asset in your asset hierarchy
            # and inform the user that it is missing !
            if orphanage_asset_external_id not in rdf_assets:
                msg = f"You dont have Orphanage asset {orphanage_asset_external_id} in asset hierarchy!"
                logging.error(msg)
                raise Exception(msg)

            logging.error("Orphaned assets will be assigned to 'Orphanage' root asset")

            for external_id in orphan_assets:
                rdf_assets[external_id]["parent_external_id"] = orphanage_asset_external_id

            orphan_assets, circular_assets = validate_asset_hierarchy(rdf_assets)

            logging.info(orphan_assets)
        else:
            logging.info("No orphaned assets found, your assets look healthy !")

        if circular_assets:
            msg = f"Found circular dependencies: {', '.join(circular_assets)}"
            logging.error(msg)
            raise Exception(msg)
        elif orphan_assets:
            msg = f"Not able to fix orphans: {', '.join(orphan_assets)}"
            logging.error(msg)
            raise Exception(msg)
        else:
            logging.info("No circular dependency among assets found, your assets hierarchy look healthy!")

        self.categorized_assets = categorize_assets(
            self.cdf_client, rdf_assets, self.dataset_id, meta_keys=self.meta_keys
        )

        count_create_assets = len(self.categorized_assets["create"])
        count_update_assets = len(self.categorized_assets["update"])
        count_decommission_assets = len(self.categorized_assets["decommission"])
        count_resurrect_assets = len(self.categorized_assets["resurrect"])

        prom_cdf_resource_stats.labels(resource_type="asset", state="create").set(count_create_assets)
        prom_cdf_resource_stats.labels(resource_type="asset", state="update").set(count_update_assets)
        prom_cdf_resource_stats.labels(resource_type="asset", state="decommission").set(count_decommission_assets)
        prom_cdf_resource_stats.labels(resource_type="asset", state="resurrect").set(count_resurrect_assets)

        self.count_create_assets = count_create_assets

        logging.info(f"Total count of assets to be created: { count_create_assets }")
        logging.info(f"Total count of assets to be updated: { count_update_assets }")
        logging.info(f"Total count of assets to be decommission: { count_decommission_assets }")
        logging.info(f"Total count of assets to be resurrect: { count_resurrect_assets }")

        msg = f"Total count of assets { len(rdf_assets) } of which: { count_create_assets } to be created"
        msg += f", { count_update_assets } to be updated"
        msg += f", { count_decommission_assets } to be decommissioned"
        msg += f", { count_resurrect_assets } to be resurrected"

        return FlowMessage(output_text=msg)

    def step_upload_cdf_assets(self, flow_msg: FlowMessage = None):
        if not self.cdf_client:
            logging.error("No CDF client available")
            raise Exception("No CDF client available")

        upload_assets(self.cdf_client, self.categorized_assets)
        for _ in range(1000):
            total_assets_after = self.cdf_client.assets.aggregate(
                filter=AssetFilter(data_set_ids=[{"id": self.dataset_id}])
            )[0]["count"]
            if total_assets_after >= self.count_create_assets:
                break
            logging.info(f"Waiting for assets to be created, current count {total_assets_after}")
            time.sleep(2)

        # UPDATE: 2023-04-05 - correct aggregation of assets in CDF for specific dataset
        total_assets_after = self.cdf_client.assets.aggregate(
            filter=AssetFilter(data_set_ids=[{"id": self.dataset_id}])
        )[0]["count"]

        prom_cdf_resource_stats.labels(resource_type="asset", state="count_after_neat_update").set(total_assets_after)
        logging.info(f"Total count of assets in CDF after update: { total_assets_after }")
        self.categorized_assets = None  # free up memory after upload .
        return FlowMessage(output_text=f"Total count of assets in CDF after update: { total_assets_after }")

    def step_prepare_cdf_relationships(self, flow_msg: FlowMessage = None):
        # create, categorize and upload relationships
        rdf_relationships = rdf2relationships(
            self.solution_graph.get_graph(),
            self.transformation_rules,
        )
        if not self.cdf_client:
            logging.info("Dry run, no CDF client available")
            return

        self.categorized_relationships = categorize_relationships(self.cdf_client, rdf_relationships, self.dataset_id)
        count_defined_relationships = len(rdf_relationships)
        count_create_relationships = len(self.categorized_relationships["create"])
        count_decommission_relationships = len(self.categorized_relationships["decommission"])
        count_resurrect_relationships = len(self.categorized_relationships["resurrect"])

        prom_cdf_resource_stats.labels(resource_type="relationships", state="defined").set(count_defined_relationships)
        prom_cdf_resource_stats.labels(resource_type="relationships", state="create").set(count_create_relationships)
        prom_cdf_resource_stats.labels(resource_type="relationships", state="decommission").set(
            count_decommission_relationships
        )
        prom_cdf_resource_stats.labels(resource_type="relationships", state="resurrect").set(
            count_resurrect_relationships
        )

        msg = (
            f"Total count of relationships { count_defined_relationships } "
            f"of which: { count_create_relationships } to be created"
        )
        msg += f", { count_decommission_relationships } to be decommissioned"
        msg += f", { count_resurrect_relationships } to be resurrected"

        return FlowMessage(output_text=msg)

    def step_upload_cdf_relationships(self, flow_msg: FlowMessage = None):
        if not self.cdf_client:
            logging.error("No CDF client available")
            raise Exception("No CDF client available")

        upload_relationships(self.cdf_client, self.categorized_relationships)

    def step_cleanup(self, flow_msg: FlowMessage):
        # TODO : cleanup
        self.categorized_assets = None
        self.categorized_relationships = None



import logging
import time
from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes import AssetFilter

from cognite.neat import rules
from cognite.neat.graph import extractors
from cognite.neat.graph.loaders.core.labels import upload_labels
from cognite.neat.graph.loaders.core.rdf_to_assets import categorize_assets, rdf2assets, upload_assets
from cognite.neat.graph.loaders.core.rdf_to_relationships import (
    categorize_relationships,
    rdf2relationships,
    upload_relationships,
)
from cognite.neat.graph.extractors import NeatGraphStore
from cognite.neat.rules.models import TransformationRules
from cognite.neat.utils.utils import add_triples
from cognite.neat.graph.loaders.validator import validate_asset_hierarchy
from cognite.neat.workflows import utils
from cognite.neat.workflows.base import BaseWorkflow
from cognite.neat.workflows.model import FlowMessage
from cognite.neat.workflows.cdf_store import CdfStore
from cognite.neat.rules.exporter import rules2graph_sheet


class SmeGraphCaptureBaseWorkflow(BaseWorkflow):
    def __init__(self, name: str, client: CogniteClient):
        super().__init__(name, client, [])
        self.counter = 0
        self.metrics.register_metric("counter_1", "", "counter", ["step"])
        self.metrics.register_metric("gauge_1", "", "gauge", ["step"])
        self.transformation_rules: TransformationRules = None
        self.solution_graph: NeatGraphStore = None
        self.dataset_id: int = 0
        self.upload_meta_object = None

    def step_cleanup(self, flow_msg: FlowMessage = None):
        logging.info("Cleanup")

    def step_error_handler(self, flow_msg: FlowMessage = None):
        logging.info("Error handler")
        return FlowMessage(output_text="Error handleed")

    # ----------------------------------------------------------------------------------
    # ------------------ Graph Capturing Sheet Generation ------------------------------
    # ----------------------------------------------------------------------------------

    def step_load_transformation_rules(self, flow_msg: FlowMessage = None):
        # Load rules from file or remote location
        self.upload_meta_object = flow_msg.payload
        cdf_store = CdfStore(self.cdf_client, self.dataset_id, rules_storage_path=self.rules_storage_path)

        rules_file = self.get_config_item_value("rules.file")
        rules_file_path = Path(self.rules_storage_path, rules_file)
        version = self.get_config_item_value("rules.version")

        if rules_file_path.exists() and not version:
            logging.info(f"Loading rules from {rules_file_path}")
        elif rules_file_path.exists() and version:
            hash = utils.get_file_hash(rules_file_path)
            if hash != version:
                cdf_store.load_rules_file_from_cdf(rules_file, version)
        else:
            cdf_store.load_rules_file_from_cdf(self.cdf_client, version)

        self.transformation_rules = rules.parse_rules_from_excel_file(rules_file_path)
        self.dataset_id = self.transformation_rules.metadata.data_set_id
        logging.info(f"Loaded prefixes {str(self.transformation_rules.prefixes)} rules")
        output_text = f"Loaded {len(self.transformation_rules.properties)} rules"
        logging.info(output_text)
        return FlowMessage(output_text=output_text)

    def step_configure_graph_store(self, flow_msg: FlowMessage = None):
        logging.info("Configure graph store")
        self.solution_graph = extractors.NeatGraphStore(prefixes=self.transformation_rules.prefixes)
        self.solution_graph.init_graph("memory")
        return FlowMessage(output_text="Graph store configured")

    def step_generate_graph_capture_sheet(self, flow_msg: FlowMessage = None):
        logging.info("Generate graph capture sheet")
        sheet_name = self.get_config_item_value("graph_capture.file", "graph_capture_sheet.xlsx")
        auto_identifier_type = self.get_config_item_value("graph_capture_sheet.auto_identifier_type", None)
        logging.info(f"Auto identifier type {auto_identifier_type}")
        data_capture_sheet_path = self.rules_storage_path.parent / "graph-sheets" / sheet_name

        rules2graph_sheet(self.transformation_rules, data_capture_sheet_path, auto_identifier_type=auto_identifier_type)

        output_text = (
            "Data capture sheet generated and can be downloaded here : "
            f'<a href="http://localhost:8000/data/graph-sheets/{sheet_name}?{time.time()}" target="_blank">'
            f"{sheet_name}</a>"
        )

        return FlowMessage(output_text=output_text)

    # ----------------------------------------------------------------------------------
    # ------------------ Graph Capturing Sheet Processing ------------------------------
    # ----------------------------------------------------------------------------------

    def step_process_graph_capture_sheet(self, flow_msg: FlowMessage = None):
        data_capture_sheet_path = Path(self.upload_meta_object["full_path"])
        logging.info(f"Processing data capture sheet {data_capture_sheet_path}")

        triples = extractors.extract_graph_from_sheet(data_capture_sheet_path, self.solution_graph)
        add_triples(self.solution_graph, triples)
        return FlowMessage(output_text="Data capture sheet processed")

    def step_create_cdf_labels(self, flow_msg: FlowMessage = None):
        logging.info("Creating CDF labels")
        upload_labels(self.cdf_client, self.transformation_rules, extra_labels=["non-historic", "historic"])

    def step_prepare_cdf_assets(self, flow_msg: FlowMessage):
        # export graph into CDF
        # TODO : decide on error handling and retry logic\

        rdf_asset_dicts = rdf2assets(
            self.solution_graph,
            self.transformation_rules,
        )

        if not self.cdf_client:
            logging.info("Dry run, no CDF client available")
            return

        # UPDATE: 2023-04-05 - correct aggregation of assets in CDF for specific dataset
        total_assets_before = self.cdf_client.assets.aggregate(
            filter=AssetFilter(data_set_ids=[{"id": self.dataset_id}])
        )[0]["count"]

        logging.info(f"Total count of assets in CDF before upload: { total_assets_before }")

        orphan_assets, circular_assets = validate_asset_hierarchy(rdf_asset_dicts)

        if orphan_assets:
            logging.error(f"Found orphaned assets: {', '.join(orphan_assets)}")

            orphanage_asset_external_id = (
                f"{self.transformation_rules.metadata.externalIdPrefix}orphanage"
                if self.transformation_rules.metadata.externalIdPrefix
                else "orphanage"
            )

            # Kill the process if you dont have orphanage asset in your asset hierarchy
            # and inform the user that it is missing !
            if orphanage_asset_external_id not in rdf_asset_dicts:
                msg = f"You dont have Orphanage asset {orphanage_asset_external_id} in asset hierarchy!"
                logging.error(msg)
                raise Exception(msg)

            logging.error("Orphaned assets will be assigned to 'Orphanage' root asset")

            for external_id in orphan_assets:
                rdf_asset_dicts[external_id]["parent_external_id"] = orphanage_asset_external_id

            orphan_assets, circular_assets = validate_asset_hierarchy(rdf_asset_dicts)

            logging.info(orphan_assets)
        else:
            logging.info("No orphaned assets found, your assets look healthy !")

        if circular_assets:
            msg = f"Found circular dependencies: {', '.join(circular_assets)}"
            logging.error(msg)
            raise Exception(msg)
        elif orphan_assets:
            msg = f"Not able to fix orphans: {', '.join(orphan_assets)}"
            logging.error(msg)
            raise Exception(msg)
        else:
            logging.info("No circular dependency among assets found, your assets hierarchy look healthy !")

        self.categorized_assets = categorize_assets(self.cdf_client, rdf_asset_dicts, self.dataset_id)

        self.asset_count = {
            "create": len(self.categorized_assets["create"]),
            "update": len(self.categorized_assets["update"]),
            "decommission": len(self.categorized_assets["decommission"]),
            "resurrect": len(self.categorized_assets["resurrect"]),
        }

        for action, value in self.asset_count.items():
            logging.info(f"Total assets to {action}: {value}")

        msg = (
            f"Total count of assets { len(rdf_asset_dicts) } of which: { self.asset_count['create'] } to be created"
            f", { self.asset_count['update'] } to be updated"
            f", { self.asset_count['decommission'] } to be decommissioned"
            f", { self.asset_count['resurrect'] } to be resurrected"
        )
        return FlowMessage(output_text=msg)

    def step_upload_cdf_assets(self, flow_msg: FlowMessage = None):
        if not self.cdf_client:
            logging.error("No CDF client available")
            raise Exception("No CDF client available")

        upload_assets(self.cdf_client, self.categorized_assets)

        for _ in range(1000):
            total_assets_after = self.cdf_client.assets.aggregate(
                filter=AssetFilter(data_set_ids=[{"id": self.dataset_id}])
            )[0]["count"]
            if total_assets_after >= self.asset_count["create"]:
                break
            logging.info(f"Waiting for assets to be created, current count {total_assets_after}")
            time.sleep(2)

        # UPDATE: 2023-04-05 - correct aggregation of assets in CDF for specific dataset
        total_assets_after = self.cdf_client.assets.aggregate(
            filter=AssetFilter(data_set_ids=[{"id": self.dataset_id}])
        )[0]["count"]

        logging.info(f"Total count of assets in CDF after update: { total_assets_after }")
        self.categorized_assets = None  # free up memory after upload .
        return FlowMessage(output_text=f"Total count of assets in CDF after update: { total_assets_after }")

    def step_prepare_cdf_relationships(self, flow_msg: FlowMessage = None):
        # create, categorize and upload relationships
        rdf_relationships = rdf2relationships(
            self.solution_graph.get_graph(),
            self.transformation_rules,
        )
        if not self.cdf_client:
            logging.info("Dry run, no CDF client available")
            return

        self.categorized_relationships = categorize_relationships(self.cdf_client, rdf_relationships, self.dataset_id)

        self.relationship_count = {
            "create": len(self.categorized_relationships["create"]),
            "decommission": len(self.categorized_relationships["decommission"]),
            "resurrect": len(self.categorized_relationships["resurrect"]),
        }

        for action, value in self.relationship_count.items():
            logging.info(f"Total assets to {action}: {value}")

        msg = (
            f"Total count of reletions { len(rdf_relationships) } of which: "
            f"{ self.relationship_count['create'] } to be created"
            f", { self.asset_count['decommission'] } to be decommissioned"
            f", { self.asset_count['resurrect'] } to be resurrected"
        )

        return FlowMessage(output_text=msg)

    def step_upload_cdf_relationships(self, flow_msg: FlowMessage = None):
        if not self.cdf_client:
            logging.error("No CDF client available")
            raise Exception("No CDF client available")

        upload_relationships(self.cdf_client, self.categorized_relationships)



from contextlib import asynccontextmanager
import logging

from fastapi import FastAPI

from cognite.neat.app.api.configuration import neat_app


@asynccontextmanager
async def lifespan(app_ref: FastAPI):
    logging.info("Startup FastAPI server")
    neat_app.set_http_server(app_ref)
    neat_app.start()
    yield
    logging.info("FastApi shutdown event")
    neat_app.stop()



from pathlib import Path

from cognite.client import CogniteClient

from cognite.neat.graph.stores import NeatGraphStore
from cognite.neat.rules.models import TransformationRules
from cognite.neat.workflows.steps.step_model import DataContract
from cognite.client.data_classes import Relationship, RelationshipUpdate


class RulesData(DataContract):
    rules: TransformationRules

    @property
    def dataset_id(self) -> int:
        return self.rules.metadata.data_set_id


class PathData(DataContract):
    excel_file_path: Path


class SourceGraph(DataContract):
    graph: NeatGraphStore


class SolutionGraph(DataContract):
    graph: NeatGraphStore


class ClientData(DataContract):
    client: CogniteClient


class CategorizedAssets(DataContract):
    assets: tuple[dict, dict] | dict


class CategorizedRelationships(DataContract):
    relationships: (
        tuple[dict[str, list[Relationship | RelationshipUpdate]], dict[str, set]]
        | dict[str, list[Relationship | RelationshipUpdate]]
    )



query_templates = [
    {"name": "Get all classes", "query": "SELECT DISTINCT ?class WHERE { ?s a ?class }"},
    {
        "name": "Get all classes with stats",
        "query": "SELECT ?class (count(?s) as ?instances ) "
        "WHERE { ?s rdf:type ?class . } group by ?class order by DESC(?instances)",
    },
    {"name": "Describe object", "query": " DESCRIBE <http://cog.com/neat#_30b297b4-8e19-da40-9f52-fb9175136a22>"},
    {"name": "Count all classes", "query": " SELECT (COUNT(DISTINCT ?class) AS ?count) WHERE { ?s a ?class }"},
    {"name": "Get all instances", "query": " SELECT DISTINCT ?s ?class WHERE { ?s a ?class }"},
    {"name": "Count all instances", "query": " SELECT (COUNT(DISTINCT ?s) AS ?count) WHERE { ?s a ?class }"},
    {"name": "Get all properties", "query": " SELECT DISTINCT ?p WHERE { ?s ?p ?o }"},
    {"name": "Count all properties", "query": " SELECT (COUNT(DISTINCT ?p) AS ?count) WHERE { ?s ?p ?o }"},
    {
        "name": "Get all properties of a class",
        "query": " SELECT DISTINCT ?property WHERE { ?s a ?class . ?s ?property ?o }",
    },
    {
        "name": "Get all properties of an instance",
        "query": " SELECT DISTINCT ?property WHERE { ?instance a ?class . ?instance ?property ?o }",
    },
    {
        "name": "Get all properties of a class and their values",
        "query": " SELECT DISTINCT ?property ?value WHERE { ?s a ?class . ?s ?property ?value }",
    },
    {
        "name": "Get all properties of an instance and their values",
        "query": " SELECT DISTINCT ?property ?value WHERE { ?instance a ?class . ?instance ?property ?value }",
    },
    {
        "name": "Get all properties of a class and their values (with type)",
        "query": """ SELECT DISTINCT ?property ?value ?type
            WHERE { ?s a ?class . ?s ?property ?value . ?value a ?type }""",
    },
    {
        "name": "Get all properties of an instance and their values (with type)",
        "query": """ SELECT DISTINCT ?property ?value ?type
            WHERE {?instance a ?class . ?instance ?property ?value . ?value a ?type }""",
    },
    {
        "name": "Get all properties of a class and their values (with type and label)",
        "query": """ SELECT DISTINCT ?property ?value ?type ?label
                      WHERE { ?s a ?class . ?s ?property ?value . ?value a ?type . ?value rdfs:label ?label }""",
    },
    {
        "name": "Get all properties of an instance and their values (with type and label)",
        "query": """ SELECT DISTINCT ?property ?value ?type ?label
            WHERE { ?instance a ?class . ?instance ?property ?value . ?value a ?type . ?value rdfs:label ?label }""",
    },
    {
        "name": "Get all properties of a class and their values (with type and label) (with type and label)",
        "query": """ SELECT DISTINCT ?property ?value ?type ?label
            WHERE { ?s a ?class . ?s ?property ?value . ?value a ?type . ?value rdfs:label ?label }""",
    },
    {
        "name": "Get all properties of an instance and their values (with type and label) (with type and label)",
        "query": """ SELECT DISTINCT ?property ?value ?type ?label
            WHERE { ?instance a ?class . ?instance ?property ?value . ?value a ?type . ?value rdfs:label ?label }""",
    },
    {
        "name": "Get list of all properties of Substations",
        "query": "SELECT ?subject ?p ?object "
        "WHERE { ?subject rdf:type cim:Substation . ?subject ?p ?object . } "
        "order by ?subject limit 12",
    },
    {
        "name": "Get all properties of the node",
        "query": "SELECT ?predicate ?object "
        "WHERE { <http://neat-cog.com/#_fbf1e5dc-2ce7-ec2a-e040-1e828c9489bf> ?predicate ?object . }",
    },
    {
        "name": "Get all properties of the node and performing 2 level traversal",
        "query": "SELECT ?property1 ?value1 ?property2 ?value2 WHERE "
        "{ <http://neat-cog.com/#_fbf1e5dc-2ce7-ec2a-e040-1e828c9489bf> ?property1 ?value1 . "
        "OPTIONAL { ?value1 ?property2 ?value2 } }",
    },
    {
        "name": "Get graph compatible query",
        "query": """
SELECT (?parentName AS ?node_name)  (?parentClass AS ?node_class) ?parentPath (?parentInst AS ?node_id )
 (?parentInst AS ?src_object_ref) (?parentInst2 AS ?dst_object_ref) WHERE {
 ?tagInst rdf:type neat:AttributeTag .
 ?tagInst neat:Path ?tagPath .
 ?tagInst neat:Value ?tagValue .
 ?tagInst neat:hasParent+ ?parentInst .
 ?parentInst neat:Name ?parentName .
 ?parentInst neat:Path ?parentPath .
 ?parentInst rdf:type ?parentClass .
 ?parentInst neat:hasParent ?parentInst2
  }  limit 100""",
    },
]



import logging
from pathlib import Path
from fastapi import APIRouter
from fastapi import HTTPException
from fastapi.responses import FileResponse
from cognite.neat.app.api.configuration import neat_app
from cognite.neat.app.api.data_classes.rest import DownloadFromCdfRequest, RunWorkflowRequest, UploadToCdfRequest
from cognite.neat.workflows import WorkflowFullStateReport
from cognite.neat.workflows.base import WorkflowDefinition
from cognite.neat.workflows.migration.wf_manifests import migrate_wf_manifest
from cognite.neat.workflows.model import FlowMessage

router = APIRouter()


@router.post("/api/workflow/start")
def start_workflow(request: RunWorkflowRequest):
    logging.info("Starting workflow endpoint")
    start_status = neat_app.workflow_manager.start_workflow_instance(
        request.name, sync=request.sync, flow_msg=FlowMessage()
    )
    result = {"workflow_instance": None, "is_success": start_status.is_success, "status_text": start_status.status_text}
    return {"result": result}


@router.get("/api/workflow/stats/{workflow_name}", response_model=WorkflowFullStateReport)
def get_workflow_stats(
    workflow_name: str,
) -> WorkflowFullStateReport | None:
    logging.info("Hit the get_workflow_stats endpoint")
    workflow = neat_app.workflow_manager.get_workflow(workflow_name)
    if workflow is None:
        raise HTTPException(status_code=404, detail="workflow not found")
    return workflow.get_state()


@router.get("/api/workflow/workflows")
def get_workflows():
    return {"workflows": neat_app.workflow_manager.get_list_of_workflows()}


@router.post("/api/workflow/create")
def create_new_workflow(request: WorkflowDefinition):
    new_workflow_definition = neat_app.workflow_manager.create_new_workflow(
        request.name, request.description, "manifest"
    )
    return {"workflow": new_workflow_definition}


@router.delete("/api/workflow/{workflow_name}")
def delete_workflow(workflow_name: str):
    neat_app.workflow_manager.delete_workflow(workflow_name)
    return {"result": "ok"}


@router.get("/api/workflow/executions")
def get_list_of_workflow_executions():
    return {"executions": neat_app.cdf_store.get_list_of_workflow_executions_from_cdf()}


@router.get("/api/workflow/detailed-execution-report/{execution_id}")
def get_detailed_execution(execution_id: str):
    return {"report": neat_app.cdf_store.get_detailed_workflow_execution_report_from_cdf(execution_id)}


@router.post("/api/workflow/reload-workflows")
def reload_workflows():
    neat_app.workflow_manager.load_workflows_from_storage()
    neat_app.triggers_manager.reload_all_triggers()
    return {"result": "ok", "workflows": neat_app.workflow_manager.get_list_of_workflows()}


@router.get("/api/workflow/workflow-definition/{workflow_name}")
def get_workflow_definition(workflow_name: str):
    workflow = neat_app.workflow_manager.get_workflow(workflow_name)
    return {"definition": workflow.get_workflow_definition()}


@router.get("/api/workflow/workflow-src/{workflow_name}/{file_name}")
def get_workflow_src(workflow_name: str, file_name: str):
    src = neat_app.workflow_manager.get_workflow_src(workflow_name, file_name=file_name)
    return FileResponse(src, media_type="text/plain")


@router.post("/api/workflow/workflow-definition/{workflow_name}")
def update_workflow_definition(workflow_name: str, request: WorkflowDefinition):
    neat_app.workflow_manager.update_workflow(workflow_name, request)
    neat_app.workflow_manager.save_workflow_to_storage(workflow_name)
    return {"result": "ok"}


@router.post("/api/workflow/upload-wf-to-cdf/{workflow_name}")
def upload_workflow_to_cdf(workflow_name: str, request: UploadToCdfRequest):
    neat_app.cdf_store.save_workflow_to_cdf(
        workflow_name, changed_by=request.author, comments=request.comments, tag=request.tag
    )
    return {"result": "ok"}


@router.post("/api/workflow/upload-rules-cdf/{workflow_name}")
def upload_rules_to_cdf(workflow_name: str, request: UploadToCdfRequest):
    file_path = Path(neat_app.config.rules_store_path, request.file_name)
    neat_app.cdf_store.save_resource_to_cdf(
        workflow_name, "neat-wf-rules", file_path, changed_by=request.author, comments=request.comments
    )
    return {"result": "ok"}


@router.post("/api/workflow/download-wf-from-cdf")
def download_wf_from_cdf(request: DownloadFromCdfRequest):
    neat_app.cdf_store.load_workflows_from_cdf(request.file_name, request.version)
    return {"result": "ok"}


@router.post("/api/workflow/download-rules-from-cdf")
def download_rules_to_cdf(request: DownloadFromCdfRequest):
    neat_app.cdf_store.load_rules_file_from_cdf(request.file_name, request.version)
    return {"file_name": request.file_name, "hash": request.version}


@router.post("/api/workflow/migrate-workflow")
def migrate_workflow():
    return migrate_wf_manifest(neat_app.config.data_store_path)


@router.get("/api/workflow/pre-cdf-assets/{workflow_name}")
def get_pre_cdf_assets(workflow_name: str):
    workflow = neat_app.workflow_manager.get_workflow(workflow_name)
    if workflow is None:
        return {"assets": []}
    return {"assets": workflow.data["CategorizedAssets"]}


@router.get("/api/workflow/context/{workflow_name}")
def get_context(workflow_name: str):
    workflow = neat_app.workflow_manager.get_workflow(workflow_name)
    context = workflow.get_context()
    objects_in_context = []
    for key, value in context.items():
        objects_in_context.append({"name": key, "type": type(value).__name__})
    return {"context": objects_in_context}


@router.get("/api/workflow/registered-steps")
def get_steps():
    steps_registry = neat_app.workflow_manager.get_steps_registry()
    return {"steps": steps_registry.get_list_of_steps()}



from cognite.neat.graph.extractors import rdf_to_graph
from cognite.neat.graph.extractors.graph_sheet_to_graph import (
    extract_graph_from_sheet,
    read_graph_excel_file_to_table_by_name,
)

__all__ = [
    "rdf_to_graph",
    "extract_graph_from_sheet",
    "read_graph_excel_file_to_table_by_name",
]



from pathlib import Path
from rdflib import Namespace
from cognite import neat


PACKAGE_DIRECTORY = Path(neat.__file__).parent


EXAMPLE_RULES = PACKAGE_DIRECTORY / "rules" / "examples"
EXAMPLE_GRAPHS = PACKAGE_DIRECTORY / "graph" / "examples"
EXAMPLE_WORKFLOWS = PACKAGE_DIRECTORY / "workflows" / "examples"


PREFIXES = {
    # "rdf": Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#"), # noqa
    # "rdfs": Namespace("http://www.w3.org/2000/01/rdf-schema#"),# noqa
    # "owl": Namespace("http://www.w3.org/2002/07/owl#"),# noqa
    # "sh": Namespace("http://www.w3.org/ns/shacl#"),
    "dct": Namespace("http://purl.org/dc/terms/"),
    "skos": Namespace("http://www.w3.org/2004/02/skos/core#"),
    "pav": Namespace("http://purl.org/pav/"),
    "cim": Namespace("http://iec.ch/TC57/2013/CIM-schema-cim16#"),
    "icim": Namespace("http://iec.ch/TC57/2013/CIM-schema-cim16-info#"),
    "entsoe": Namespace("http://entsoe.eu/CIM/SchemaExtension/3/1#"),
    "entsoe2": Namespace("http://entsoe.eu/CIM/SchemaExtension/3/2#"),
    "md": Namespace("http://iec.ch/TC57/61970-552/ModelDescription/1#"),
    "pti": Namespace("http://www.pti-us.com/PTI_CIM-schema-cim16#"),
    "tnt": Namespace("http://purl.org/cognite/tnt#"),
    "neat": Namespace("http://purl.org/cognite/neat#"),
}

DEFAULT_NAMESPACE = Namespace("http://purl.org/cognite/app#")
DEFAULT_URI = ""



from dataclasses import dataclass
from typing import Self
from cognite.neat.rules import _exceptions
from cognite.neat.rules._validation import (
    are_entity_names_dms_compliant,
    are_properties_redefined,
)
from cognite.neat.rules.analysis import to_class_property_pairs
from cognite.neat.rules.models import DATA_TYPE_MAPPING, TransformationRules
from cognite.neat.utils.utils import generate_exception_report


_TYPE = (
    "{% include 'type_header' %}type {{ class_definition.class_id }} {{'{'}}"
    "{%-for property_definition in class_properties%}"
    "{% include 'field' %}"
    "{% endfor %}\n}\n"
)

_TYPE_HEADER = (
    "{%- if header %}"
    "{%- if class_definition.description and class_definition.class_name %}"
    '"""\n{{class_definition.description}}\n@name {{ class_definition.class_name }}\n"""\n{##}\n'
    "{%- elif class_definition.description %}"
    '\n"""\n{{class_definition.description}}\n"""\n'
    "{%- endif %}"
    "{%- endif %}"
)


_FIELD = (
    "{% include 'field_header' %}\n"
    "  {{ property_definition.property_id }}: "
    "{%-if property_definition.property_type == 'DatatypeProperty'%}"
    "{% include 'attribute_occurrence' %}"
    "{%-else%}"
    "{% include 'edge_occurrence' %}"
    "{%- endif -%}"
)

_FIELD_HEADER = (
    "{%- if header %}"
    "{%- if property_definition.description and property_definition.property_name %}"
    '\n  """\n  {{property_definition.description}}'
    "\n  @name {{ property_definition.property_name }}"
    '\n  """\n  '
    "{%- elif property_definition.description %}"
    '\n  """\n  {{property_definition.description}}'
    '\n  """\n  '
    "{%- endif %}"
    "{%- endif %}"
)

_ATTRIBUTE_OCCURRENCE = (
    "{%-if property_definition.min_count and property_definition.max_count == 1%}"
    " {% include 'value_type' %}!"
    "{%-elif property_definition.min_count and property_definition.max_count != 1%}"
    " [{% include 'value_type' %}!]!"
    "{%-elif property_definition.max_count != 1%}"
    " [{% include 'value_type' %}!]!"
    "{%-else%}{% include 'value_type' %}"
    "{%- endif -%}"
)

_EDGE_OCCURRENCE = (
    "{%-if not(property_definition.min_count and property_definition.max_count == 1)%}"
    " [{% include 'value_type' %}]"
    "{%-else%}"
    " {% include 'value_type' %}"
    "{%- endif -%}"
)

_FIELD_VALUE_TYPE = """{{value_type_mapping[property_definition.expected_value_type]['GraphQL']
                if property_definition.expected_value_type in value_type_mapping
                else property_definition.expected_value_type}}"""


@dataclass
class GraphQLSchema:
    """Abilities to generate a GraphQL schema from TransformationRules"""

    schema: str

    @classmethod
    def from_rules(cls, transformation_rules: TransformationRules, verbose: bool = False) -> Self:
        names_compliant, name_warnings = are_entity_names_dms_compliant(transformation_rules, return_report=True)
        if not names_compliant:
            raise _exceptions.Error10(report=generate_exception_report(name_warnings))

        properties_redefined, redefinition_warnings = are_properties_redefined(transformation_rules, return_report=True)
        if properties_redefined:
            raise _exceptions.Error11(report=generate_exception_report(redefinition_warnings))

        return cls(schema=cls.generate_schema(transformation_rules, verbose))

    @staticmethod
    def generate_schema(transformation_rules: TransformationRules, verbose: bool) -> str:
        """Generates a GraphQL schema of given TransformationRules"""
        class_properties = to_class_property_pairs(transformation_rules)

        type_definitions = []

        for class_id in class_properties:
            parameters = {
                "class_definition": transformation_rules.classes[class_id],
                "class_properties": list(class_properties[class_id].values()),
                "value_type_mapping": DATA_TYPE_MAPPING,
                "header": verbose,
            }

            type_definitions.append(GraphQLSchema.template().render(parameters))

        return "\n\n".join(type_definitions)

    @staticmethod
    def template():
        from jinja2 import DictLoader, Environment, Template

        template: Template = Environment(
            loader=DictLoader(
                {
                    "type_header": _TYPE_HEADER,
                    "type": _TYPE,
                    "field_header": _FIELD_HEADER,
                    "field": _FIELD,
                    "value_type": _FIELD_VALUE_TYPE,
                    "edge_occurrence": _EDGE_OCCURRENCE,
                    "attribute_occurrence": _ATTRIBUTE_OCCURRENCE,
                }
            ),
            cache_size=1000,
        ).get_template("type")
        return template



import logging
from pathlib import Path

from rdflib import RDF, Literal, URIRef
from cognite.neat.constants import PREFIXES
from cognite.neat.graph.transformations.transformer import RuleProcessingReport, domain2app_knowledge_graph
from cognite.neat.rules import parse_rules_from_excel_file
from cognite.neat.rules.exporter.rules2triples import get_instances_as_triples
from cognite.neat.workflows import utils
from cognite.neat.workflows.cdf_store import CdfStore
from cognite.neat.workflows.model import FlowMessage, WorkflowConfigItem
from cognite.neat.workflows.steps.step_model import Step

from cognite.client import CogniteClient
from ..data_contracts import RulesData, SolutionGraph, SourceGraph

__all__ = [
    "LoadTransformationRules",
    "TransformSourceToSolutionGraph",
    "LoadInstancesFromRulesToSolutionGraph",
    "LoadDataModelFromRulesToSourceGraph",
]


class LoadTransformationRules(Step):
    description = "The step loads transformation rules from the file or remote location"
    category = "rules"
    configuration_templates = [
        WorkflowConfigItem(
            name="rules.file",
            value="rules.xlsx",
            label="Full name of the rules file",
        ),
        WorkflowConfigItem(name="rules.version", value="", label="Optional version of the rules file"),
    ]

    def run(self, cdf_store: CdfStore) -> (FlowMessage, RulesData):
        rules_file = self.configs.get_config_item_value("rules.file")
        rules_file_path = Path(self.data_store_path, "rules", rules_file)
        version = self.configs.get_config_item_value("rules.version", default_value=None)
        if not rules_file_path.exists():
            logging.info(f"Rules files doesn't exist in local fs {rules_file_path}")

        if rules_file_path.exists() and not version:
            logging.info(f"Loading rules from {rules_file_path}")
        elif rules_file_path.exists() and version:
            hash = utils.get_file_hash(rules_file_path)
            if hash != version:
                cdf_store.load_rules_file_from_cdf(rules_file, version)
        else:
            cdf_store.load_rules_file_from_cdf(rules_file, version)

        transformation_rules = parse_rules_from_excel_file(rules_file_path)
        rules_metrics = self.metrics.register_metric(
            "data_model_rules", "Transformation rules stats", m_type="gauge", metric_labels=["component"]
        )
        rules_metrics.labels({"component": "classes"}).set(len(transformation_rules.classes))
        rules_metrics.labels({"component": "properties"}).set(len(transformation_rules.properties))
        logging.info(f"Loaded prefixes {str(transformation_rules.prefixes)} rules from {rules_file_path.name!r}.")
        output_text = f"Loaded {len(transformation_rules.properties)} rules"
        return FlowMessage(output_text=output_text), RulesData(rules=transformation_rules)


class TransformSourceToSolutionGraph(Step):
    description = "The step transforms source graph to solution graph"
    category = "transformation"

    def run(
        self,
        transformation_rules: RulesData,
        cdf_client: CogniteClient,
        source_graph: SourceGraph,
        solution_graph: SolutionGraph,
    ) -> FlowMessage:
        solution_graph.graph.drop()
        report = RuleProcessingReport()
        # run transformation and generate new graph
        solution_graph.graph.set_graph(
            domain2app_knowledge_graph(
                source_graph.graph.get_graph(),
                transformation_rules.rules,
                app_instance_graph=solution_graph.graph.get_graph(),
                extra_triples=get_instances_as_triples(transformation_rules.rules),
                client=cdf_client,
                cdf_lookup_database=None,  # change this accordingly!
                processing_report=report,
            )
        )
        return FlowMessage(
            output_text=f"Total processed rules: { report.total_rules } , success: { report.total_success } , \
             no results: { report.total_success_no_results } , failed: { report.total_failed }",
            payload=report,
        )


class LoadInstancesFromRulesToSolutionGraph(Step):
    description = "The step loads instances from rules file into solution graph."
    category = "graph_loader"

    def run(self, transformation_rules: RulesData, solution_graph: SolutionGraph) -> FlowMessage:
        triples = get_instances_as_triples(transformation_rules.rules)
        instance_ids = {triple[0] for triple in triples}
        output_text = f"Loaded {len(instance_ids)} instances out of"

        try:
            for triple in triples:
                solution_graph.graph.graph.add(triple)
        except Exception as e:
            logging.error("Not able to load instances to source graph")
            raise e

        output_text = f"Loaded {len(triples)} statements defining"
        output_text += f" {len(instance_ids)} instances"
        return FlowMessage(output_text=output_text)


class LoadDataModelFromRulesToSourceGraph(Step):
    description = "The step loads data model from rules file into source graph."
    category = "graph_loader"

    def run(self, transformation_rules: RulesData, source_graph: SourceGraph) -> FlowMessage:
        ns = PREFIXES["neat"]
        clases = transformation_rules.rules.classes
        properties = transformation_rules.rules.properties
        counter = 0
        for class_name, class_def in clases.items():
            rdf_instance_id = URIRef(ns + "_" + class_def.class_id)
            source_graph.graph.graph.add((rdf_instance_id, URIRef(ns + "Name"), Literal(class_name)))
            source_graph.graph.graph.add((rdf_instance_id, RDF.type, URIRef(ns + class_def.class_id)))
            if class_def.parent_class:
                source_graph.graph.graph.add(
                    (rdf_instance_id, URIRef(ns + "hasParent"), URIRef(ns + "_" + class_def.parent_class))
                )
            counter += 1

        for property_name, property_def in properties.items():
            rdf_instance_id = URIRef(ns + "_" + property_def.class_id)
            source_graph.graph.graph.add(
                (rdf_instance_id, URIRef(ns + property_def.property_id), Literal(property_def.expected_value_type))
            )
            if property_def.expected_value_type not in ("string", "integer", "float", "boolean"):
                source_graph.graph.graph.add(
                    (rdf_instance_id, URIRef(ns + "connectedTo"), URIRef(ns + "_" + property_def.expected_value_type))
                )
            counter += 1

        output_text = f"Loaded {counter} classes into source graph"
        return FlowMessage(output_text=output_text)



import logging
from pathlib import Path
import time
from cognite.neat.workflows.model import FlowMessage, WorkflowConfigItem
from cognite.neat.graph import extractors
from cognite.neat.workflows.steps.step_model import Step
from ..data_contracts import RulesData, SolutionGraph
from cognite.neat.rules.exporter import rules2graph_sheet
from cognite.neat.utils.utils import add_triples

__all__ = ["GenerateDataCaptureSpreadsheet", "ProcessDataCaptureSpreadsheetIntoSolutionGraph"]


class GenerateDataCaptureSpreadsheet(Step):
    description = "The step generates data capture spreadsheet from data model defined in rules"
    category = "data_capture"
    configuration_templates = [
        WorkflowConfigItem(
            name="graph_capture.file",
            value="graph_capture_sheet.xlsx",
            label="File name of the data capture sheet",
        ),
        WorkflowConfigItem(
            name="graph_capture_sheet.auto_identifier_type", value="index-based", label="Type of automatic identifier"
        ),
        WorkflowConfigItem(
            name="graph_capture_sheet.storage_dir", value="staging", label="Directory to store data capture sheets"
        ),
    ]

    def run(self, rules: RulesData) -> FlowMessage:
        logging.info("Generate graph capture sheet")
        sheet_name = self.configs.get_config_item_value("graph_capture.file", "graph_capture_sheet.xlsx")
        auto_identifier_type = self.configs.get_config_item_value("graph_capture_sheet.auto_identifier_type", None)
        staging_dir_str = self.configs.get_config_item_value("graph_capture_sheet.storage_dir", "staging")
        logging.info(f"Auto identifier type {auto_identifier_type}")
        staging_dir = self.data_store_path / Path(staging_dir_str)
        staging_dir.mkdir(parents=True, exist_ok=True)
        data_capture_sheet_path = staging_dir / sheet_name

        rules2graph_sheet.rules2graph_capturing_sheet(
            rules.rules, data_capture_sheet_path, auto_identifier_type=auto_identifier_type
        )

        output_text = (
            "Data capture sheet generated and can be downloaded here : "
            f'<a href="http://localhost:8000/data/{staging_dir_str}/{sheet_name}?{time.time()}" target="_blank">'
            f"{sheet_name}</a>"
        )
        return FlowMessage(output_text=output_text)


class ProcessDataCaptureSpreadsheetIntoSolutionGraph(Step):
    description = "The step processes data capture spreadsheet into solution graph"
    category = "data_capture"

    def run(
        self,
        transformation_rules: RulesData,
        solution_graph: SolutionGraph,
    ) -> FlowMessage:
        triggered_flow_message = self.flow_context["StartFlowMessage"]
        data_capture_sheet_path = Path(triggered_flow_message.payload["full_path"])
        logging.info(f"Processing data capture sheet {data_capture_sheet_path}")

        triples = extractors.extract_graph_from_sheet(
            data_capture_sheet_path, transformation_rule=transformation_rules.rules
        )
        add_triples(solution_graph.graph, triples)
        return FlowMessage(output_text="Data capture sheet processed")



"""
"""
import re
from collections import Counter
from dataclasses import dataclass
from enum import StrEnum
from typing import Literal, Optional, Self

from pydantic import BaseModel, field_validator

from . import _exceptions


@dataclass
class Triple:
    subject: str
    predicate: str
    object: str = None
    optional: bool = False


class OWL(StrEnum):
    class_ = "class"
    property_ = "property"
    object_property = "object_property"
    data_property = "data_property"
    annotation_property = "annotation_property"


class RuleType(StrEnum):
    rdfpath = "rdfpath"
    rawlookup = "rawlookup"
    sparql = "sparql"


class Lookup(StrEnum):
    table = "table"
    key = "key"
    value = "value"  # type: ignore


# here word can be any alphanumeric character, dash, underscore or dot
# which are used to define prefixes and entity names
# this causes super long processing time for some reason

_prefix = r"[a-zA-Z]+[a-zA-Z0-9]*[-_.]*[a-zA-Z0-9]+"
_name = r"[a-zA-Z0-9-_.]+[a-zA-Z0-9]+"


# entity can be either anything except literal value!
_entity = rf"{_prefix}:{_name}"

entity = re.compile(rf"^(?P<prefix>{_prefix}):(?P<name>{_name})$")

# traversal direction
_direction = r"(?P<direction>(->|<-))"

# we are defining classes and their properties
# though we have implicit way of defining object properties and that is using -> or <- operators
_class = rf"(?P<{OWL.class_}>{_entity})"
_property = rf"\((?P<{OWL.property_}>{_entity})\)"


_class_only = re.compile(rf"^{_class}$")

_step = rf"((->|<-){_class}({_property})?)"
_steps = re.compile(_step)
_step_class_only = re.compile(rf"(^{_direction}{_class})$")
_step_class_and_property = re.compile(rf"(^{_direction}{_class}{_property}$)")


_traversal = "traversal"
_origin = rf"(?P<origin>{_entity})"

_hop = re.compile(rf"^{_origin}(?P<{_traversal}>{_step}+)$")

# grabbing specific property for a class, property can be either object, annotation or data property
_single_property = re.compile(rf"^{_class}{_property}$")

# grabbing all properties for a class
_all_properties = re.compile(rf"^{_class}\(\*\)$")

_all_traversal = rf"({_class}\(\*\)|{_class}{_property}|{_origin}(?P<{_traversal}>{_step}+))"

_table = re.compile(rf"^(?P<{Lookup.table}>{_name})\((?P<{Lookup.key}>{_name}),\s*(?P<{Lookup.value}>{_name})\)$")


class Entity(BaseModel):
    """Entity is a class or property in OWL/RDF sense."""

    prefix: str
    name: str

    @property
    def id(self) -> str:
        return f"{self.prefix}:{self.name}"

    def __repr__(self):
        return self.id

    @classmethod
    def from_string(cls, raw: str, **kwargs) -> Self:
        if result := entity.match(raw):
            return cls(prefix=result.group("prefix"), name=result.group("name"), **kwargs)
        else:
            raise ValueError(f"{cls.__name__} is expected to be prefix:name, got {raw}")

    @classmethod
    def from_list(cls, prefix, names: list[str]) -> list[Self]:
        # TODO: Prefixes should be a list of prefixes as you might instances of classes
        # from different namespaces. This is a case when a given data model reuses classes
        # from different ontologies/data models, which is in case of Statnett.
        return [cls(prefix=prefix, name=name) for name in names]


StepDirection = Literal["source", "target", "origin"]
_direction_by_symbol = {"->": "target", "<-": "source"}


class Step(BaseModel):
    class_: Entity
    property: Optional[Entity] = None  # only terminal step has property
    direction: StepDirection

    @classmethod
    def from_string(cls, raw: str, **kwargs) -> Self:
        if result := _step_class_and_property.match(raw):
            return cls(
                class_=Entity.from_string(result.group(OWL.class_)),
                property=Entity.from_string(result.group(OWL.property_)),
                direction=_direction_by_symbol[result.group("direction")],
                **kwargs,
            )
        elif result := _step_class_only.match(raw):
            return cls(
                class_=Entity.from_string(result.group(OWL.class_)),
                direction=_direction_by_symbol[result.group("direction")],
            )  # type: ignore
        msg = f"Invalid step {raw}, expected in one of the following forms:"
        msg += " ->prefix:name, <-prefix:name, ->prefix:name(prefix:name) or <-prefix:name(prefix:name)"
        raise ValueError(msg)


class Traversal(BaseModel):
    pass


class SingleProperty(Traversal):
    class_: Entity
    property: Entity

    @field_validator("class_", "property", mode="before")
    def process_if_string(cls, value):
        return Entity.from_string(value) if isinstance(value, str) else value


class AllReferences(Traversal):
    class_: Entity

    @field_validator("class_", mode="before")
    def process_if_string(cls, value):
        return Entity.from_string(value) if isinstance(value, str) else value


class AllProperties(Traversal):
    class_: Entity

    @field_validator("class_", mode="before")
    def process_if_string(cls, value):
        return Entity.from_string(value) if isinstance(value, str) else value


class Origin(BaseModel):
    class_: Entity

    @field_validator("class_", mode="before")
    def process_if_string(cls, value):
        return Entity.from_string(value) if isinstance(value, str) else value


class Hop(Traversal):
    """Multi or single hop traversal through graph"""

    class_: Entity
    traversal: list[Step]

    @field_validator("class_", mode="before")
    def process_origin_if_string(cls, value):
        return Entity.from_string(value) if isinstance(value, str) else value

    @field_validator("traversal", mode="before")
    def process_path_if_string(cls, value):
        if isinstance(value, str):
            return [Step.from_string(result[0]) for result in _steps.findall(value)]
        return value


class TableLookup(BaseModel):
    name: str
    key: str
    value: str


class Rule(BaseModel):
    pass


class RDFPath(Rule):
    traversal: Traversal


class RawLookup(RDFPath):
    table: TableLookup


class Query(BaseModel):
    query: str


class SPARQLQuery(Rule):
    traversal: Query


def parse_traversal(raw: str) -> AllReferences | AllProperties | SingleProperty | Hop:
    if result := _class_only.match(raw):
        return AllReferences(class_=result.group(OWL.class_))
    elif result := _all_properties.match(raw):
        return AllProperties(class_=result.group(OWL.class_))
    elif result := _single_property.match(raw):
        return SingleProperty(class_=result.group(OWL.class_), property=result.group(OWL.property_))
    elif result := _hop.match(raw):
        return Hop(class_=result.group("origin"), traversal=result.group(_traversal))
    else:
        raise _exceptions.Error1(raw).to_pydantic_custom_error()


def parse_table_lookup(raw: str) -> TableLookup:
    if result := _table.match(raw):
        return TableLookup(
            name=result.group(Lookup.table), key=result.group(Lookup.key), value=result.group(Lookup.value)
        )
    raise _exceptions.Error2(raw).to_pydantic_custom_error()


def parse_rule(rule_raw: str, rule_type: RuleType) -> RDFPath:
    match rule_type:
        case RuleType.rdfpath:
            rule_raw = rule_raw.replace(" ", "")
            return RDFPath(traversal=parse_traversal(rule_raw))
        case RuleType.rawlookup:
            rule_raw = rule_raw.replace(" ", "")
            if Counter(rule_raw).get("|") != 1:
                raise _exceptions.Error3(rule_raw).to_pydantic_custom_error()
            traversal, table_lookup = rule_raw.split("|")
            return RawLookup(traversal=parse_traversal(traversal), table=parse_table_lookup(table_lookup))
        case RuleType.sparql:
            return SPARQLQuery(traversal=Query(query=rule_raw))


def is_valid_rule(rule_type: RuleType, rule_raw: str) -> bool:
    is_valid_rule = {RuleType.rdfpath: is_rdfpath, RuleType.rawlookup: is_rawlookup}[rule_type]
    return is_valid_rule(rule_raw)


def is_rdfpath(raw: str) -> bool:
    try:
        parse_traversal(raw)
    except ValueError:
        return False
    return True


def is_rawlookup(raw: str) -> bool:
    try:
        parse_rule(raw, RuleType.rawlookup)
    except ValueError:
        return False
    return True



"""This module is used to generate mock graph data for purposes of testing of NEAT.
"""
import logging
import warnings
from collections import OrderedDict

import pandas as pd
from prometheus_client import Gauge
from rdflib import RDF, Literal, Namespace, URIRef

from cognite.neat.rules.analysis import (
    get_class_linkage,
    get_classes_with_properties,
    get_defined_classes,
    get_symmetric_pairs,
)
from cognite.neat.rules.exporter.rules2rules import subset_rules
from cognite.neat.rules.models import TransformationRules
from cognite.neat.utils.utils import remove_namespace

neat_total_processed_mock_triples = Gauge(
    "neat_total_processed_mock_triples", "Number of processed mock triples", ["state"]
)
neat_mock_triples_processing_timing = Gauge(
    "neat_mock_triples_processing_timing", "Generation of mock knowledge graph timing metrics", ["aggregation"]
)


def generate_triples(
    transformation_rules: TransformationRules,
    class_count: dict,
    stop_on_exception: bool = False,
    allow_isolated_classes: bool = True,
) -> list[tuple]:
    """Generate mock triples for purposes of testing of NEAT

    Parameters
    ----------
    data_model : DataModelingDefinition
        Data model containing ontology and object shape constraints for each class
    class_count : dict
        Target class count for each class in the ontology
    stop_on_exception: bool
        To stop if exception is encountered or not, default is False
    allow_isolated_classes: bool
        To allow generation of instances for classes that are not connected to any other class, default is True
    """

    # Figure out which classes are defined in the data model and which are not

    defined_classes = get_defined_classes(transformation_rules)

    if non_existing_classes := set(class_count.keys()) - defined_classes:
        msg = f"Class count contains classes {non_existing_classes} for which properties are not defined in Data Model!"
        if stop_on_exception:
            logging.error(msg)
            raise ValueError(msg)
        else:
            msg += " These classes will be ignored."
            logging.warning(msg)
            warnings.warn(msg, stacklevel=2)
            for class_ in non_existing_classes:
                class_count.pop(class_)

    # Subset data model to only classes that are defined in class count
    transformation_rules = (
        subset_rules(transformation_rules, set(class_count.keys()), skip_validation=True)
        if defined_classes != set(class_count.keys())
        else transformation_rules
    )

    class_linkage = get_class_linkage(transformation_rules)

    # Remove one of symmetric pairs from class linkage to maintain proper linking
    # among instances of symmetrically linked classes
    if sym_pairs := get_symmetric_pairs(transformation_rules):
        class_linkage = _remove_higher_occurring_sym_pair(class_linkage, sym_pairs)

    # Remove any of symmetric pairs containing classes that are not present class count
    class_linkage = _remove_non_requested_sym_pairs(class_linkage, class_count)

    # Generate generation order for classes instances
    generation_order = _prettify_generation_order(_get_generation_order(class_linkage))

    # Generated simple view of data model
    class_definitions = _rules_to_dict(transformation_rules)

    # pregenerate instance ids for each remaining class
    instance_ids = {
        key: [URIRef(transformation_rules.metadata.namespace[f"{key}-{i}"]) for i in range(value)]
        for key, value in class_count.items()
    }

    # create triple for each class instance defining its type
    triples = []
    for class_ in class_count:
        triples += [
            (
                class_instance_id,
                RDF.type,
                URIRef(transformation_rules.metadata.namespace[class_]),
            )
            for class_instance_id in instance_ids[class_]
        ]

    # generate triples for connected classes
    for class_ in generation_order:
        triples += _generate_triples_per_class(
            class_,
            class_definitions,
            sym_pairs,
            instance_ids,
            transformation_rules.metadata.namespace,
            stop_on_exception,
        )

    # generate triples for isolated classes
    if allow_isolated_classes:
        for class_ in set(class_count.keys()) - set(generation_order):
            triples += _generate_triples_per_class(
                class_,
                class_definitions,
                sym_pairs,
                instance_ids,
                transformation_rules.metadata.namespace,
                stop_on_exception,
            )

    return triples


def _get_generation_order(
    class_linkage: pd.DataFrame, parent_col: str = "source_class", child_col: str = "target_class"
) -> dict:
    parent_child_list = class_linkage[[parent_col, child_col]].values.tolist()
    # Build a directed graph and a list of all names that have no parent
    graph = {name: set() for tup in parent_child_list for name in tup}
    has_parent = {name: False for tup in parent_child_list for name in tup}
    for parent, child in parent_child_list:
        graph[parent].add(child)
        has_parent[child] = True

    # All names that have absolutely no parent:
    roots = [name for name, parents in has_parent.items() if not parents]

    return _traverse({}, graph, roots)


def _traverse(hierarchy: dict, graph: dict, names: str) -> dict:
    """traverse the graph and return the hierarchy"""
    for name in names:
        hierarchy[name] = _traverse({}, graph, graph[name])
    return hierarchy


def _prettify_generation_order(generation_order: dict, depth: dict = None, start=-1) -> dict:
    """Prettifies generation order dictionary for easier consumption."""
    depth = depth or {}
    for key, value in generation_order.items():
        depth[key] = start + 1
        if isinstance(value, dict):
            _prettify_generation_order(value, depth, start=start + 1)
    return OrderedDict(sorted(depth.items(), key=lambda item: item[1]))


def _remove_higher_occurring_sym_pair(class_linkage: pd.DataFrame, sym_pairs: set[tuple[str, str]]) -> pd.DataFrame:
    """Remove symmetric pair which is higher in occurrence."""
    rows_to_remove = set()
    for source, target in sym_pairs:
        first_sym_property_occurrence = class_linkage[
            (class_linkage.source_class == source) & (class_linkage.target_class == target)
        ].max_occurrence.values[0]
        second_sym_property_occurrence = class_linkage[
            (class_linkage.source_class == target) & (class_linkage.target_class == source)
        ].max_occurrence.values[0]

        if first_sym_property_occurrence is None:
            # this means that source occurrence is unbounded
            index = class_linkage[
                (class_linkage.source_class == source) & (class_linkage.target_class == target)
            ].index.values[0]
        elif (
            second_sym_property_occurrence is None
            or first_sym_property_occurrence <= second_sym_property_occurrence
            and second_sym_property_occurrence > first_sym_property_occurrence
        ):
            # this means that target occurrence is unbounded
            index = class_linkage[
                (class_linkage.source_class == target) & (class_linkage.target_class == source)
            ].index.values[0]
        else:
            index = class_linkage[
                (class_linkage.source_class == source) & (class_linkage.target_class == target)
            ].index.values[0]
        rows_to_remove.add(index)

    return class_linkage.drop(rows_to_remove)


def _remove_non_requested_sym_pairs(class_linkage: pd.DataFrame, class_count: dict) -> pd.DataFrame:
    """Remove symmetric pairs which classes are not found in class count."""
    rows_to_remove = set(class_linkage[~(class_linkage["source_class"].isin(set(class_count.keys())))].index.values)
    rows_to_remove |= set(class_linkage[~(class_linkage["target_class"].isin(set(class_count.keys())))].index.values)

    return class_linkage.drop(rows_to_remove)


def _generate_mock_data_property_triples(
    instance_ids: list[URIRef], property_: str, namespace: Namespace
) -> list[tuple[URIRef, URIRef, Literal]]:
    """Generates triples for data properties."""
    # TODO: we should handle datatype, currently we are assuming everything is string
    return [
        (
            id_,
            URIRef(namespace[property_]),
            Literal(remove_namespace(id_).replace("-", " ")),
        )
        for id_ in instance_ids
    ]


def _generate_mock_object_property_triples(
    class_: str,
    property_definition: pd.Series,
    class_definitions: dict[str, pd.DataFrame],
    sym_pairs: set[tuple[str, str]],
    instance_ids: list[URIRef],
    namespace: Namespace,
    stop_on_exception: bool,
) -> list[tuple[URIRef, URIRef, URIRef]]:
    """Generates triples for object properties."""
    if property_definition.value_type not in instance_ids:
        msg = f"Class {property_definition.value_type} not found in class count! "
        if stop_on_exception:
            logging.error(msg)
            raise ValueError(msg)
        else:
            msg += (
                f"Skipping creating triples for property {property_definition.name} "
                f"of class {class_} which expects values of this type!"
            )
            logging.warning(msg)
            warnings.warn(msg, stacklevel=2)
            return []

    # Handling symmetric property
    symmetric_property = (
        class_definitions[property_definition.value_type][
            class_definitions[property_definition.value_type].value_type == class_
        ].index.values[0]
        if tuple((class_, property_definition.value_type)) in sym_pairs
        else None
    )

    triples = []

    for i, source in enumerate(instance_ids[class_]):
        target = instance_ids[property_definition.value_type][i % len(instance_ids[property_definition.value_type])]
        triples += [
            (
                URIRef(source),
                URIRef(namespace[property_definition.name]),
                URIRef(target),
            )
        ]

        if symmetric_property:
            triples += [
                (
                    URIRef(target),
                    URIRef(namespace[symmetric_property]),
                    URIRef(source),
                )
            ]

    # remove symmetric property from class definition of downstream class
    # to avoid asymmetric linking in mock graph
    if symmetric_property:
        class_definitions[property_definition.value_type].drop(symmetric_property, inplace=True)

    return triples


def _generate_triples_per_class(
    class_: str,
    class_definitions: dict[str, pd.DataFrame],
    sym_pairs: set[tuple[str, str]],
    instance_ids: dict[str, list[URIRef]],
    namespace: Namespace,
    stop_on_exception: bool,
) -> list[tuple]:
    """Generate triples for a given class."""
    triples = []
    for _, property_definition in class_definitions[class_].iterrows():
        if property_definition.property_type == "DatatypeProperty":
            triples += _generate_mock_data_property_triples(instance_ids[class_], property_definition.name, namespace)

        elif property_definition.property_type == "ObjectProperty":
            triples += _generate_mock_object_property_triples(
                class_,
                property_definition,
                class_definitions,
                sym_pairs,
                instance_ids,
                namespace,
                stop_on_exception,
            )

        else:
            logging.error(f"Property type {property_definition.property_type} not supported!")
            raise ValueError(f"Property type {property_definition.property_type} not supported!")

    return triples


def _rules_to_dict(transformation_rules: TransformationRules) -> dict[str, pd.DataFrame]:
    """Represent data model as a dictionary of data frames, where each data frame
    represents properties defined for a given class.

    Returns
    -------
    Dict[str, pd.DataFrame]
        Simplified representation of the data model
    """

    data_model = {}

    defined_classes = get_classes_with_properties(transformation_rules)

    for class_ in defined_classes:
        properties = {}
        for property_ in defined_classes[class_]:
            if property_.property_id not in properties:
                properties[property_.property_id] = {
                    "property_type": property_.property_type,
                    "value_type": property_.expected_value_type,
                    "min_count": property_.min_count,
                    "max_count": property_.max_count,
                }

        data_model[class_] = pd.DataFrame(properties).T

    return data_model



from pathlib import Path
import time
from cognite.neat.rules.exporter.rules2graphql import GraphQLSchema
from cognite.neat.workflows.model import FlowMessage, WorkflowConfigItem
from cognite.neat.workflows.steps.data_contracts import RulesData
from cognite.neat.workflows.steps.step_model import Step

__all__ = ["GenerateGraphQLSchemaFromRules"]


class GenerateGraphQLSchemaFromRules(Step):
    description = "The step generates GraphQL schema from data model defined in rules"
    category = "fdm"
    configuration_templates = [
        WorkflowConfigItem(
            name="fdm_schema.file",
            value="fdm_model.graphql",
            label="Name of the FDM schema file",
        ),
        WorkflowConfigItem(
            name="graph_ql_export.storage_dir", value="staging", label="Directory to store FDM schema file"
        ),
    ]

    def run(self, transformation_rules: RulesData) -> FlowMessage:
        data_model_gql = GraphQLSchema.from_rules(transformation_rules.rules, verbose=True).schema

        default_name = (
            f"{transformation_rules.rules.metadata.prefix}-"
            f"v{transformation_rules.rules.metadata.version.strip().replace('.', '_')}"
            ".graphql"
        )
        schema_name = self.configs.get_config_item_value("fdm_schema.file", default_name)
        staging_dir_str = self.configs.get_config_item_value("graph_ql_export.storage_dir", "staging")
        staging_dir = self.data_store_path / Path(staging_dir_str)
        staging_dir.mkdir(parents=True, exist_ok=True)
        fdm_model_full_path = staging_dir / schema_name

        with fdm_model_full_path.open(mode="w") as fdm_file:
            fdm_file.write(data_model_gql)

        output_text = (
            "FDM Schema generated and can be downloaded here : "
            f'<a href="http://localhost:8000/data/{staging_dir_str}/{schema_name}?{time.time()}" '
            f'target="_blank">{schema_name}</a>'
        )

        return FlowMessage(output_text=output_text)



import logging
import os
import shutil
import time
from pathlib import Path

import pandas as pd
import pyoxigraph
import requests
from prometheus_client import Gauge, Summary
from rdflib import Graph, Namespace
from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore
from rdflib.query import Result

from cognite.neat.constants import PREFIXES, DEFAULT_NAMESPACE
from cognite.neat.graph.stores.configuration import RdfStoreType
from cognite.neat.graph.stores import oxrdflib
from cognite.neat.graph.extractors.rdf_to_graph import rdf_file_to_graph

prom_qsm = Summary("store_query_time_summary", "Time spent processing queries", ["query"])
prom_sq = Gauge("store_single_query_time", "Time spent processing a single query", ["query"])


class NeatGraphStore:
    """NeatGraphStore is a class that stores the graph and provides methods to read/write data it contains


    Attributes
    ----------
    graph : Graph
        Instance of rdflib.Graph class for graph storage
    base_prefix : str
        Used as a base prefix for graph namespace, allowing querying graph data using a shortform of a URI
    namespace : Namespace
        Namespace (aka URI) used to resolve any relative URI in the graph
    prefixes : dict[str, URIRef]
        Dictionary of additional prefixes used in the graph
    """

    def __init__(
        self,
        graph: Graph = None,
        base_prefix: str = "",  # usually empty
        namespace: Namespace = DEFAULT_NAMESPACE,
        prefixes: dict = PREFIXES,
    ):
        self.graph: Graph = graph
        self.base_prefix: str = base_prefix
        self.namespace: Namespace = namespace
        self.prefixes: dict[str, Namespace] = prefixes

        self.rdf_store_type: str = RdfStoreType.MEMORY
        self.rdf_store_query_url: str = None
        self.rdf_store_update_url: str = None
        self.returnFormat: str = None
        self.df_cache: pd.DataFrame = None
        self.graph_db_rest_url: str = "http://localhost:7200"
        self.internal_storage_dir: Path = None

    def init_graph(
        self,
        rdf_store_type: str = RdfStoreType.MEMORY,
        rdf_store_query_url: str = None,
        rdf_store_update_url: str = None,
        graph_name: str = None,
        base_prefix: str = None,
        returnFormat: str = "csv",
        internal_storage_dir: Path = None,
    ):
        """Initializes the graph.

        Parameters
        ----------
        rdf_store_type : str, optional
            Graph store type, by default RdfStoreType.MEMORY
        rdf_store_query_url : str, optional
            URL towards which SPARQL query is executed, by default None
        rdf_store_update_url : str, optional
            URL towards which SPARQL update is executed, by default None
        graph_name : str, optional
            Name of graph, by default None
        base_prefix : str
            Base prefix for graph namespace to change if needed, by default None
        returnFormat : str, optional
            Transport format of graph data between, by default "csv"
        internal_storage_dir : Path, optional
            Path to directory where internal storage is located, by default None (in-memory storage).
            Used only for Oxigraph.
        """
        logging.info("Initializing NeatGraphStore")
        self.rdf_store_type = rdf_store_type
        self.rdf_store_query_url = rdf_store_query_url
        self.rdf_store_update_url = rdf_store_update_url
        self.graph_name = graph_name
        self.returnFormat = returnFormat
        self.internal_storage_dir = Path(internal_storage_dir) if internal_storage_dir else None

        if self.rdf_store_type in ["", RdfStoreType.MEMORY]:
            logging.info("Initializing graph in memory")
            self.graph = Graph()
        elif self.rdf_store_type == RdfStoreType.OXIGRAPH:
            logging.info("Initializing Oxigraph store")
            # Adding support for both in-memory and file-based storage
            oxstore = None
            self.internal_storage_dir.mkdir(parents=True, exist_ok=True)
            for i in range(4):
                try:
                    oxstore = pyoxigraph.Store(
                        path=str(self.internal_storage_dir) if self.internal_storage_dir else None
                    )  # Store (Rust object) accepts only str as path and not Path.
                    break
                except OSError as e:
                    if "lock" in str(e) and i < 3:
                        # lock originated from another instance of the store
                        logging.error("Error initializing Oxigraph store: %s", e)
                        logging.info("Removing LOCK file and retrying")
                        time.sleep(1)
                    else:
                        raise e

            self.graph = Graph(store=oxrdflib.OxigraphStore(store=oxstore))
            self.graph.default_union = True

        elif self.rdf_store_type == RdfStoreType.GRAPHDB:
            logging.info("Initializing graph store with GraphDB")
            store = SPARQLUpdateStore(
                query_endpoint=self.rdf_store_query_url,
                update_endpoint=self.rdf_store_update_url,
                returnFormat=self.returnFormat,
                context_aware=False,
                postAsEncoded=False,
                autocommit=False,
            )
            self.graph = Graph(store=store)
        else:
            logging.error("Unknown RDF store type: %s", self.rdf_store_type)
            raise Exception("Unknown RDF store type: %s", self.rdf_store_type)

        if self.prefixes:
            for prefix, namespace in self.prefixes.items():
                logging.info("Adding prefix %s with namespace %s", prefix, namespace)
                self.graph.bind(prefix, namespace)

        if base_prefix:
            self.base_prefix = base_prefix

        self.graph.bind(self.base_prefix, self.namespace)
        logging.info("Adding prefix %s with namespace %s", self.base_prefix, self.namespace)
        logging.info("Graph initialized")

    def close(self):
        """Closes the graph."""
        if self.rdf_store_type == RdfStoreType.OXIGRAPH:
            try:
                if self.graph:
                    self.graph.store._inner.flush()
                    self.graph.close()
            except Exception as e:
                logging.debug("Error closing graph: %s", e)

    def import_from_file(self, file_path: Path = None):
        """Imports graph data from file.

        Parameters
        ----------
        file_path : Path, optional
            File path to file containing graph data, by default None
        """

        if not file_path:
            file_path = self.config.rdf_import_path

        if self.rdf_store_type == RdfStoreType.OXIGRAPH:
            self.graph.store._inner.bulk_load(str(file_path), "application/rdf+xml", base_iri=self.namespace)
            self.graph.store._inner.optimize()
        else:
            self.graph = rdf_file_to_graph(file_path, base_namespace=self.namespace, prefixes=self.prefixes)
        return

    def get_graph(self) -> Graph:
        """Returns the graph."""
        return self.graph

    def set_graph(self, graph: Graph):
        """Sets the graph."""
        self.graph = graph

    def query(self, query: str) -> Result:
        """Returns the result of the query."""
        logging.info(f"Contexts: {list(self.graph.store.contexts())}")
        start_time = time.perf_counter()
        result = self.graph.query(query)
        stop_time = time.perf_counter()
        elapsed_time = stop_time - start_time
        prom_qsm.labels("query").observe(elapsed_time)
        prom_sq.labels("query").set(elapsed_time)
        return result

    def drop(self):
        """Drops the graph."""
        if self.rdf_store_type == RdfStoreType.MEMORY:
            # In case of in-memory graph, we just reinitialize the graph
            # otherwise we would lose the prefixes and bindings which fails
            # workflow
            self.init_graph(
                self.rdf_store_type,
                self.rdf_store_query_url,
                self.rdf_store_update_url,
                self.graph_name,
                self.base_prefix,
                self.returnFormat,
            )
        if self.rdf_store_type == RdfStoreType.OXIGRAPH:
            try:
                files_before = os.listdir(self.internal_storage_dir)
                self.graph.store._inner.flush()
                self.graph.store._inner.clear()
                time.sleep(1)  # delay to allow the store finish file operations
                files_after = os.listdir(self.internal_storage_dir)
                new_files = set(files_after) - set(files_before)
                # static_files = ["CURRENT", "IDENTITY", "LOCK"]
                static_files = [f for f in files_before if f.startswith("MANIFEST") or f.startswith("OPTIONS")]
                static_files.extend(["CURRENT", "IDENTITY", "LOCK"])

                files_to_keep = set(static_files) | new_files
                logging.info(f"Files to keep: {files_to_keep}")
                for f in os.listdir(self.internal_storage_dir):
                    if f not in files_to_keep:
                        try:
                            os.remove(os.path.join(self.internal_storage_dir, f))
                        except Exception as e:
                            logging.error(f"Error deleting file {f}: {e}")
            except Exception as e:
                logging.error(f"Error dropping graph : {e}")

        elif self.rdf_store_type == RdfStoreType.GRAPHDB:
            r = requests.delete(f"{self.graph_db_rest_url}/repositories/{self.graph_name}/rdf-graphs/service?default")
            logging.info(
                f"Dropped graph with state: {r.text}",
            )

    def query_to_dataframe(
        self,
        query: str,
        column_mapping: dict = None,
        save_to_cache: bool = False,
        index_column: str = "instance",
    ) -> pd.DataFrame:
        """Returns the result of the query as a dataframe.

        Parameters
        ----------
        query : str
            SPARQL query to execute
        column_mapping : dict, optional
            Columns name mapping, by default None
        save_to_cache : bool, optional
            Save result of query to cache, by default False
        index_column : str, optional
            Indexing column , by default "instance"

        Returns
        -------
        pd.DataFrame
            Dataframe with result of query
        """

        if column_mapping is None:
            column_mapping = {0: "instance", 1: "property", 2: "value"}

        result = self.graph.query(query, DEBUG=False)
        df_cache = pd.DataFrame(list(result))
        df_cache.rename(columns=column_mapping, inplace=True)
        df_cache[index_column] = df_cache[index_column].apply(lambda x: str(x))
        if save_to_cache:
            self.df_cache = df_cache
        return df_cache

    def __del__(self):
        if self.rdf_store_type == RdfStoreType.OXIGRAPH:
            if self.graph:
                if self.graph.store:
                    self.graph.store._inner.flush()
                self.graph.close()
            # It requires more investigation os.remove(self.internal_storage_dir / "LOCK")

    def get_df(self) -> pd.DataFrame:
        """Returns the cached dataframe."""
        return self.df_cache

    def get_instance_properties_from_cache(self, instance_id: str) -> pd.DataFrame:
        """Returns the properties of an instance."""
        return self.df_cache.loc[self.df_cache["instance"] == instance_id]

    def print_triples(self):
        """Prints the triples of the graph."""
        for subj, pred, obj in self.graph:
            logging.info(f"Triple: {subj} {pred} {obj}")


def drop_graph_store(graph: NeatGraphStore, storage_path: Path, force: bool = False):
    """Drops graph store by flushing in-flight data , releasing locks and completely
    removing all files the storage path."""
    if graph:
        if graph.rdf_store_type == RdfStoreType.OXIGRAPH and storage_path:
            if storage_path.exists():
                graph.__del__()
                graph = None
                # remove all files in the storage path except the lock file
                for f in os.listdir(storage_path):
                    # if f != "LOCK":
                    os.remove(os.path.join(storage_path, f))
                logging.info("Graph store dropped.")
            else:
                logging.info(f"Storage path {storage_path} does not exist. Skipping drop.")
        else:
            logging.info(
                "Graph store {graph.rdf_store_type} is not Oxigraph or storage path is not provided. Skipping drop."
            )
    elif force and storage_path:
        if storage_path.exists():
            shutil.rmtree(storage_path)



from cognite.client import CogniteClient

from cognite.neat.workflows.model import FlowMessage


class WorkflowTaskBuilder:
    """Collection of all base tasks for workflows.All tasks must run in the context of a workflow including threads."""

    def __init__(self, cdf_client: CogniteClient, workflow_manager):
        # TODO : figure out circular import and set type to WorkflowManager
        self.cdf_client = cdf_client
        self.workflow_manager = workflow_manager

    def start_workflow_task(self, workflow_name: str, sync: bool, flow_message: FlowMessage):
        """Call a workflow task from another workflow"""
        return self.workflow_manager.start_workflow_instance(
            workflow_name=workflow_name, flow_msg=flow_message, sync=sync
        )



import logging
from pathlib import Path

from cognite.client import CogniteClient

from cognite.neat import rules
from cognite.neat.graph import extractors
from cognite.neat.constants import PREFIXES
from cognite.neat.graph.extractors import NeatGraphStore, drop_graph_store
from cognite.neat.rules.exporter.rules2triples import get_instances_as_triples
from cognite.neat.rules.models import TransformationRules
from cognite.neat.graph.transformations.transformer import RuleProcessingReport, domain2app_knowledge_graph
from cognite.neat.workflows import utils
from cognite.neat.workflows.base import BaseWorkflow
from cognite.neat.workflows.model import FlowMessage
from cognite.neat.workflows.cdf_store import CdfStore


class GraphsAndRulesBaseWorkflow(BaseWorkflow):
    def __init__(self, name: str, client: CogniteClient):
        super().__init__(name, client, [])
        self.dataset_id: int = 0
        self.source_graph: NeatGraphStore = None
        self.solution_graph: NeatGraphStore = None
        self.transformation_rules: TransformationRules = None
        self.graph_source_type = "memory"

    def step_load_transformation_rules(self, flow_msg: FlowMessage = None):
        # Load rules from file or remote location
        cdf_store = CdfStore(self.cdf_client, self.dataset_id, rules_storage_path=self.rules_storage_path)

        rules_file = self.get_config_item("rules.file").value
        rules_file_path = Path(self.rules_storage_path, rules_file)
        version = self.get_config_item("rules.version").value

        if rules_file_path.exists() and not version:
            logging.info(f"Loading rules from {rules_file_path}")
        elif rules_file_path.exists() and version:
            hash = utils.get_file_hash(rules_file_path)
            if hash != version:
                cdf_store.load_rules_file_from_cdf(rules_file, version)
        else:
            cdf_store.load_rules_file_from_cdf(self.cdf_client, version)

        self.transformation_rules = rules.parse_rules_from_excel_file(rules_file_path)
        self.dataset_id = self.transformation_rules.metadata.data_set_id
        logging.info(f"Loaded prefixes {str(self.transformation_rules.prefixes)} rules from {rules_file_path.name!r}.")
        output_text = f"Loaded {len(self.transformation_rules.properties)} rules"
        logging.info(output_text)
        return FlowMessage(output_text=output_text)

    def step_configuring_stores(self, flow_msg: FlowMessage = None, clean_start: bool = True):
        # Initialize source and solution graph stores . clean_start=True will delete all
        # artifacts(files , locks , etc) from previous runs
        logging.info("Initializing source graph")
        self.graph_source_type = self.get_config_item_value("source_rdf_store.type", self.graph_source_type)
        source_store_dir = self.get_config_item_value("source_rdf_store.disk_store_dir", None)
        solution_store_dir = self.get_config_item_value("solution_rdf_store.disk_store_dir", None)
        source_store_dir = Path(self.data_store_path) / Path(source_store_dir) if source_store_dir else None
        solution_store_dir = Path(self.data_store_path) / Path(solution_store_dir) if solution_store_dir else None
        logging.info(f"source_store_dir={source_store_dir}")
        logging.info(f"solution_store_dir={solution_store_dir}")
        if clean_start:
            drop_graph_store(self.source_graph, source_store_dir, force=True)
            drop_graph_store(self.solution_graph, solution_store_dir, force=True)

        self.source_graph = extractors.NeatGraphStore(
            prefixes=self.transformation_rules.prefixes, base_prefix="neat", namespace=PREFIXES["neat"]
        )

        if self.get_config_item_value("source_rdf_store.type"):
            self.source_graph.init_graph(
                self.get_config_item_value("source_rdf_store.type", self.graph_source_type),
                self.get_config_item_value("source_rdf_store.query_url"),
                self.get_config_item_value("source_rdf_store.update_url"),
                "neat-tnt",
                internal_storage_dir=source_store_dir,
            )

        if self.get_config_item_value("solution_rdf_store.type"):
            self.solution_graph = extractors.NeatGraphStore(
                prefixes=self.transformation_rules.prefixes, base_prefix="neat", namespace=PREFIXES["neat"]
            )

            self.solution_graph.init_graph(
                self.get_config_item_value("solution_rdf_store.type"),
                self.get_config_item_value("solution_rdf_store.query_url"),
                self.get_config_item_value("solution_rdf_store.update_url"),
                "tnt-solution",
                internal_storage_dir=solution_store_dir,
            )

        self.solution_graph.graph_db_rest_url = self.get_config_item_value("solution_rdf_store.api_root_url")
        return

    def step_load_source_graph(self, flow_msg: FlowMessage = None):
        # Load graph into memory or GraphDB
        if self.graph_source_type.lower() == "graphdb":
            try:
                result = self.source_graph.query("SELECT DISTINCT ?class WHERE { ?s a ?class }")
            except Exception as e:
                logging.error(f"Failed to query most likely remote DB is not running {e}")
                raise Exception("Failed to query graph , most likely remote DB is not running") from e
            else:
                logging.info(f"Loaded {len(result.bindings)} classes")
        elif self.graph_source_type.lower() in ("memory", "oxigraph"):
            if source_file := self.get_config_item_value("source_rdf_store.file"):
                graphs = Path(self.data_store_path) / "source-graphs"
                self.source_graph.import_from_file(graphs / source_file)
                logging.info(f"Loaded {source_file} into source graph.")
            else:
                raise ValueError("You need a source_rdf_store.file specified for source_rdf_store.type=memory")
        else:
            raise NotImplementedError(f"Graph type {self.graph_source_type} is not supported.")

        self.solution_graph.drop()
        return

    def step_run_transformation(self, flow_msg: FlowMessage = None):
        report = RuleProcessingReport()
        # run transformation and generate new graph
        self.solution_graph.set_graph(
            domain2app_knowledge_graph(
                self.source_graph.get_graph(),
                self.transformation_rules,
                app_instance_graph=self.solution_graph.get_graph(),
                extra_triples=get_instances_as_triples(self.transformation_rules),
                client=self.cdf_client,
                cdf_lookup_database=None,  # change this accordingly!
                processing_report=report,
            )
        )
        return FlowMessage(
            output_text=f"Total processed rules: { report.total_rules } , success: { report.total_success } , \
             no results: { report.total_success_no_results } , failed: { report.total_failed }",
            payload=report,
        )



from __future__ import annotations

import math
import re
import warnings
from datetime import datetime
from pathlib import Path
from typing import Any, ClassVar, Dict, Optional

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    HttpUrl,
    ValidationError,
    constr,
    field_validator,
    model_validator,
    parse_obj_as,
    validator,
)
from pydantic.fields import FieldInfo
from rdflib import XSD, Literal, Namespace, URIRef

from cognite.client.data_classes.data_modeling.data_types import (
    Timestamp,
    Text,
    Int32,
    Int64,
    Float32,
    Boolean,
    TimeSeriesReference,
    FileReference,
    SequenceReference,
    Json,
)

from cognite.neat.constants import PREFIXES

# from . import _exceptions
from cognite.neat.rules import _exceptions
from cognite.neat.rules.to_rdf_path import (
    AllReferences,
    Entity,
    Hop,
    RawLookup,
    RuleType,
    SPARQLQuery,
    SingleProperty,
    Traversal,
    parse_rule,
)

__all__ = ["Class", "Instance", "Metadata", "Prefixes", "Property", "Resource", "TransformationRules"]

# mapping of XSD types to Python and GraphQL types
DATA_TYPE_MAPPING = {
    "boolean": {"python": bool, "GraphQL": "Boolean", "dms": Boolean},
    "float": {"python": float, "GraphQL": "Float", "dms": Float32},
    "integer": {"python": "int", "GraphQL": "Int", "dms": Int32},
    "nonPositiveInteger": {"python": int, "GraphQL": "Int", "dms": Int32},
    "nonNegativeInteger": {"python": int, "GraphQL": "Int", "dms": Int32},
    "negativeInteger": {"python": "int", "GraphQL": "Int", "dms": Int32},
    "long": {"python": int, "GraphQL": "Int", "dms": Int64},
    "string": {"python": str, "GraphQL": "String", "dms": Text},
    "anyURI": {"python": str, "GraphQL": "String", "dms": Text},
    "normalizedString": {"python": str, "GraphQL": "String", "dms": Text},
    "token": {"python": str, "GraphQL": "String", "dms": Text},
    # Graphql does not have a datetime type this is CDF specific
    "dateTime": {"python": datetime, "GraphQL": "Timestamp", "dms": Timestamp},
    # CDF specific types, not in XSD
    "timeseries": {"python": TimeSeriesReference, "GraphQL": "TimeSeries", "dms": TimeSeriesReference},
    "file": {"python": FileReference, "GraphQL": "File", "dms": FileReference},
    "sequence": {"python": SequenceReference, "GraphQL": "Sequence", "dms": TimeSeriesReference},
    "json": {"python": Json, "GraphQL": "Json", "dms": Json},
}


def type_to_target_convention(type_: str, target_type_convention: str) -> type:
    """Returns the GraphQL type for a given XSD type."""
    return DATA_TYPE_MAPPING[type_][target_type_convention]


METADATA_VALUE_MAX_LENGTH = 5120


def replace_nan_floats_with_default(values: dict, model_fields: dict[str, FieldInfo]) -> dict:
    output = {}
    for field_name, value in values.items():
        is_nan_float = isinstance(value, float) and math.isnan(value)
        if not is_nan_float:
            output[field_name] = value
            continue
        if field_name in model_fields:
            output[field_name] = model_fields[field_name].default
        else:
            # field_name may be an alias
            source_name = next((name for name, field in model_fields.items() if field.alias == field_name), None)
            if source_name:
                output[field_name] = model_fields[source_name].default
            else:
                # Just pass it through if it is not an alias.
                output[field_name] = value
    return output


class RuleModel(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True, str_strip_whitespace=True, arbitrary_types_allowed=True, strict=False, extra="allow"
    )

    @classmethod
    def mandatory_fields(cls, use_alias=False) -> set[str]:
        """Returns a set of mandatory fields for the model."""
        return _get_required_fields(cls, use_alias)


def _get_required_fields(model: type[BaseModel], use_alias: bool = False) -> set[str]:
    """Get required fields from a pydantic model.

    Parameters
    ----------
    model : type[BaseModel]
        Pydantic data model
    use_alias : bool, optional
        Whether to return field alias name, by default False

    Returns
    -------
    list[str]
        List of required fields
    """
    required_fields = set()
    for name, field in model.model_fields.items():
        if not field.is_required():
            continue

        alias = getattr(field, "alias", None)
        if use_alias and alias:
            required_fields.add(alias)
        else:
            required_fields.add(name)
    return required_fields


class URL(BaseModel):
    url: HttpUrl


Description = constr(min_length=1, max_length=255)

# regex expressions for compliance of Metadata sheet parsing
prefix_compliance_regex = r"^([a-zA-Z]+)([a-zA-Z0-9]*[_-]{0,1}[a-zA-Z0-9])+$"
cdf_space_name_compliance_regex = rf"(?!^(space|cdf|dms|pg3|shared|system|node|edge)$)({prefix_compliance_regex})"
data_model_name_compliance_regex = r"^([a-zA-Z]+)([a-zA-Z0-9]*[_-]{0,1}[a-zA-Z0-9])+$"
version_compliance_regex = (
    r"^([0-9]+[_-]{1}[0-9]+[_-]{1}[0-9]+[_-]{1}[a-zA-Z0-9]+)|"
    r"([0-9]+[_-]{1}[0-9]+[_-]{1}[0-9]+)|([0-9]+[_-]{1}[0-9])|([0-9]+)$"
)

Prefix = constr(min_length=1, max_length=43)
ExternalId = constr(min_length=1, max_length=255)


class Metadata(RuleModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True, str_strip_whitespace=True, arbitrary_types_allowed=True
    )
    prefix: Prefix = Field(
        alias="shortName",
        description="This is used as prefix for generation of RDF OWL/SHACL data model representation",
    )
    cdf_space_name: Prefix = Field(
        description="This is used as CDF space name to which model is intend to be stored. "
        "By default it is set to 'playground'",
        alias="cdfSpaceName",
        default="playground",
    )

    namespace: Optional[Namespace] = Field(
        description="This is used as RDF namespace for generation of RDF OWL/SHACL data model representation "
        "and/or for generation of RDF graphs",
        min_length=1,
        max_length=2048,
        default=None,
    )
    data_model_name: Optional[ExternalId] = Field(
        description="Name that uniquely identifies data model",
        alias="dataModelName",
        default=None,
    )

    version: str = Field(
        min_length=1,
        max_length=43,
    )
    is_current_version: bool = Field(alias="isCurrentVersion", default=True)
    created: datetime
    updated: datetime = Field(default_factory=lambda: datetime.utcnow())
    title: str = Field(min_length=1, max_length=255)
    description: Description
    creator: str | list[str]
    contributor: Optional[str | list[str]] = None
    rights: Optional[str] = "Restricted for Internal Use of Cognite"
    externalIdPrefix: Optional[str] = Field(alias="externalIdPrefix", default=None)
    data_set_id: Optional[int] = Field(alias="dataSetId", default=None)
    source: Optional[str | Path] = Field(
        description="File path to Excel file which was used to produce Transformation Rules",
        default=None,
    )
    dms_compliant: bool = True

    @field_validator(
        "externalIdPrefix",
        "contributor",
        "contributor",
        "description",
        "rights",
        mode="before",
    )
    def replace_float_nan_with_default(cls, value, info):
        if isinstance(value, float) and math.isnan(value):
            return cls.model_fields[info.field_name].default
        return value

    @field_validator("version", mode="before")
    def convert_to_string(cls, value):
        return str(value)

    @validator("prefix", always=True)
    def is_prefix_compliant(cls, value):
        if not re.match(prefix_compliance_regex, value):
            raise _exceptions.Error100(value, prefix_compliance_regex).to_pydantic_custom_error()
        else:
            return value

    @validator("cdf_space_name", always=True)
    def is_cdf_space_name_compliant(cls, value):
        if not re.match(cdf_space_name_compliance_regex, value):
            raise _exceptions.Error101(value, cdf_space_name_compliance_regex).to_pydantic_custom_error()
        else:
            return value

    @validator("namespace", always=True)
    def set_namespace_if_none(cls, value, values):
        if value is None:
            if values["cdf_space_name"] == "playground":
                return Namespace(f"http://purl.org/cognite/{values['prefix']}#")
            else:
                return Namespace(f"http://purl.org/cognite/{values['cdf_space_name']}/{values['prefix']}#")
        try:
            return Namespace(parse_obj_as(HttpUrl, value))
        except ValidationError:
            raise _exceptions.Error102(value).to_pydantic_custom_error()

    @validator("namespace", always=True)
    def fix_namespace_ending(cls, value):
        if value.endswith("#") or value.endswith("/"):
            return value
        warnings.warn(_exceptions.Warning100(value).message, category=_exceptions.Warning100, stacklevel=2)
        return f"{value}#"

    @validator("data_model_name", always=True)
    def set_data_model_name_if_none(cls, value, values):
        if value is not None:
            return value
        warnings.warn(
            _exceptions.Warning101(values["prefix"].replace("-", "_")).message,
            category=_exceptions.Warning101,
            stacklevel=2,
        )
        return values["prefix"].replace("-", "_")

    @validator("data_model_name", always=True)
    def is_data_model_name_compliant(cls, value):
        if not re.match(data_model_name_compliance_regex, value):
            raise _exceptions.Error103(value, data_model_name_compliance_regex).to_pydantic_custom_error()
        else:
            return value

    @validator("version", always=True)
    def is_version_compliant(cls, value):
        # turn "." into "_" to avoid issues with CDF
        if "." in value:
            warnings.warn(_exceptions.Warning102().message, category=_exceptions.Warning102, stacklevel=2)
            value = value.replace(".", "_")
        if not re.match(version_compliance_regex, value):
            raise _exceptions.Error104(value, version_compliance_regex).to_pydantic_custom_error()
        else:
            return value

    @field_validator("creator", "contributor", mode="before")
    def to_list_if_comma(cls, value, info):
        if isinstance(value, str):
            if value:
                return value.replace(", ", ",").split(",")
            if cls.model_fields[info.field_name].default is None:
                return None
        return value


class Resource(RuleModel):
    # Solution model
    description: Optional[Description] = Field(alias="Description", default=None)

    # Solution CDF resource, it is not needed when working with FDM, this is only for
    # Classic CDF data model
    cdf_resource_type: Optional[str] = Field(alias="Resource Type", default=None)

    # Advance data modeling: Keeping track if Resource got deprecated or not
    deprecated: bool = Field(default=False)
    deprecation_date: Optional[datetime] = Field(alias="deprecationDate", default=None)
    replaced_by: Optional[str] = Field(alias="replacedBy", default=None)

    # Advance data modeling: Relation to existing resources for purpose of mapping
    source: Optional[HttpUrl] = Field(
        alias="Source", description="Source of information for given entity, e.g. CIM", default=None
    )
    source_entity_name: Optional[str] = Field(
        alias="Source Entity Name", description="Closest entity in source, e.g. Substation", default=None
    )
    match_type: Optional[str] = Field(
        alias="Match Type", description="Type of match between source entity and one being defined", default=None
    )
    comment: Optional[str] = Field(alias="Comment", description="Comment about mapping", default=None)

    @model_validator(mode="before")
    def replace_float_nan_with_default(cls, values: dict) -> dict:
        return replace_nan_floats_with_default(values, cls.model_fields)


class_id_compliance_regex = r"^([a-zA-Z]+)([a-zA-Z0-9]+[._-]{0,1}[a-zA-Z0-9._-]+)+$"


class Class(Resource):
    class_id: ExternalId = Field(
        alias="Class",
    )
    class_name: Optional[ExternalId] = Field(alias="Name", default=None)
    # Solution model
    parent_class: Optional[ExternalId] = Field(alias="Parent Class", default=None)

    # Solution CDF resource
    parent_asset: Optional[ExternalId] = Field(alias="Parent Asset", default=None)

    @model_validator(mode="before")
    def replace_nan_floats_with_default(cls, values: dict) -> dict:
        return replace_nan_floats_with_default(values, cls.model_fields)

    @validator("class_id", always=True)
    def is_class_id_compliant(cls, value):
        if not re.match(class_id_compliance_regex, value):
            raise _exceptions.Error200(value, class_id_compliance_regex).to_pydantic_custom_error()
        else:
            return value

    @validator("class_name", always=True)
    def set_class_name_if_none(cls, value, values):
        if value is None:
            if "class_id" not in values:
                raise _exceptions.Error201().to_pydantic_custom_error()
            warnings.warn(
                _exceptions.Warning200(values["class_id"]).message, category=_exceptions.Warning200, stacklevel=2
            )
            value = values["class_id"]
        return value


property_id_compliance_regex = r"^(\*)|(([a-zA-Z]+)([a-zA-Z0-9]+[._-]{0,1}[a-zA-Z0-9._-]+)+)$"


class Property(Resource):
    # Solution model
    class_id: ExternalId = Field(alias="Class")
    property_id: ExternalId = Field(alias="Property")
    property_name: Optional[ExternalId] = Field(alias="Name", default=None)
    expected_value_type: ExternalId = Field(alias="Type")
    min_count: Optional[int] = Field(alias="Min Count", default=0)
    max_count: Optional[int] = Field(alias="Max Count", default=None)
    default: Any = Field(None)

    # OWL property
    property_type: str = "DatatypeProperty"

    # Solution CDF resource
    resource_type_property: Optional[list[str]] = Field(alias="Resource Type Property", default=None)
    source_type: str = Field(alias="Relationship Source Type", default="Asset")
    target_type: str = Field(alias="Relationship Target Type", default="Asset")
    label: Optional[str] = Field(alias="Relationship Label", default=None)
    relationship_external_id_rule: Optional[str] = Field(alias="Relationship ExternalID Rule", default=None)

    # Transformation rule (domain to solution)
    rule_type: Optional[RuleType] = Field(alias="Rule Type", default=None)
    rule: Optional[str | AllReferences | SingleProperty | Hop | RawLookup | SPARQLQuery | Traversal] = Field(
        alias="Rule", default=None
    )
    skip_rule: bool = Field(alias="Skip", default=False)
    mandatory: bool = False

    # Specialization of cdf_resource_type to allow definition of both
    # Asset and Relationship at the same time
    cdf_resource_type: list[str] = Field(alias="Resource Type", default=[])

    @property
    def is_raw_lookup(self) -> bool:
        return self.rule_type == RuleType.rawlookup

    @model_validator(mode="before")
    def replace_float_nan_with_default(cls, values: dict) -> dict:
        return replace_nan_floats_with_default(values, cls.model_fields)

    @validator("class_id", always=True)
    def is_class_id_compliant(cls, value):
        if not re.match(class_id_compliance_regex, value):
            raise _exceptions.Error300(value, class_id_compliance_regex).to_pydantic_custom_error()
        else:
            return value

    @validator("property_id", always=True)
    def is_property_id_compliant(cls, value):
        if not re.match(property_id_compliance_regex, value):
            raise _exceptions.Error301(value, property_id_compliance_regex).to_pydantic_custom_error()
        else:
            return value

    @validator("expected_value_type", always=True)
    def is_expected_value_type_compliant(cls, value):
        if not re.match(class_id_compliance_regex, value):
            raise _exceptions.Error302(value, class_id_compliance_regex).to_pydantic_custom_error()
        else:
            return value

    @validator("rule_type", pre=True)
    def to_lowercase(cls, value):
        return value.casefold() if value else value

    @validator("skip_rule", pre=True)
    def from_string(cls, value):
        if isinstance(value, str):
            return value.casefold() in {"true", "skip", "yes", "y"}
        return value

    @validator("rule")
    def is_valid_rule(cls, value, values):
        if rule_type := values.get("rule_type"):
            if not value:
                raise _exceptions.Error305(
                    values["property_id"], values["class_id"], values["rule_type"]
                ).to_pydantic_custom_error()
            _ = parse_rule(value, rule_type)
        return value

    @validator("resource_type_property", pre=True)
    def split_str(cls, v):
        if v:
            return [v.strip() for v in v.split(",")] if "," in v else [v]

    @field_validator("cdf_resource_type", mode="before")
    def to_list_if_comma(cls, value, info):
        if isinstance(value, str):
            if value:
                return value.replace(", ", ",").split(",")
            if cls.model_fields[info.field_name].default is None:
                return None
        return value

    # Setters
    # TODO: configure setters to only run if field_validators are successful, otherwise do not run them!
    @property
    def is_mandatory(self) -> bool:
        return self.min_count != 0

    @model_validator(mode="after")
    def set_property_type(self):
        if self.expected_value_type in DATA_TYPE_MAPPING.keys():
            self.property_type = "DatatypeProperty"
        else:
            self.property_type = "ObjectProperty"
        return self

    @model_validator(mode="after")
    def set_property_name_if_none(self):
        if self.property_name is None:
            warnings.warn(
                _exceptions.Warning300(self.property_id).message, category=_exceptions.Warning300, stacklevel=2
            )
            self.property_name = self.property_id
        return self

    @model_validator(mode="after")
    def set_relationship_label(self):
        if self.label is None:
            warnings.warn(
                _exceptions.Warning301(self.property_id).message, category=_exceptions.Warning301, stacklevel=2
            )
            self.label = self.property_id
        return self

    @model_validator(mode="after")
    def set_skip_rule(self):
        if self.rule_type is None:
            warnings.warn(
                _exceptions.Warning302(class_id=self.class_id, property_id=self.property_id).message,
                category=_exceptions.Warning302,
                stacklevel=2,
            )
            self.skip_rule = True
        else:
            self.skip_rule = False
        return self


class Prefixes(RuleModel):
    prefixes: Dict[str, Namespace] = PREFIXES


class Instance(RuleModel):
    """Class deals with instances of classes in the data model"""

    instance: Optional[URIRef] = Field(alias="Instance", default=None)
    property_: Optional[URIRef] = Field(alias="Property", default=None)
    value: Optional[Literal | URIRef] = Field(alias="Value", default=None)
    namespace: Namespace
    prefixes: Dict[str, Namespace]

    @staticmethod
    def get_value(value, prefixes) -> URIRef | Literal:
        try:
            url = URL(url=value).url
            return URIRef(url)
        except ValidationError:
            try:
                entity = Entity.from_string(value)
                return URIRef(prefixes[entity.prefix][entity.name])
            except ValueError:
                return value

    @model_validator(mode="before")
    def convert_values(cls, values: dict):
        # we expect to read Excel sheet which contains naming convention of column
        # 'Instance', 'Property', 'Value', if that's not the case we should raise error
        if not {"Instance", "Property", "Value"}.issubset(set(values.keys())):
            raise TypeError("We only support inputs from the transformation rule Excel sheet!!!")

        namespace = values["namespace"]
        prefixes = values["prefixes"]

        values["Instance"] = cls.get_value(values["Instance"], prefixes)
        values["Instance"] = (
            values["Instance"] if isinstance(values["Instance"], URIRef) else URIRef(namespace[values["Instance"]])
        )

        values["Property"] = cls.get_value(values["Property"], prefixes)
        values["Property"] = (
            values["Property"] if isinstance(values["Property"], URIRef) else URIRef(namespace[values["Property"]])
        )

        if isinstance(values["Value"], str):
            values["Value"] = cls.get_value(values["Value"], prefixes)
            if not isinstance(values["Value"], URIRef):
                datatype = (
                    XSD.integer
                    if cls.isint(values["Value"])
                    else XSD.float
                    if cls.isfloat(values["Value"])
                    else XSD.string
                )
                values["Value"] = Literal(values["Value"], datatype=datatype)
        elif isinstance(values["Value"], float):
            values["Value"] = Literal(values["Value"], datatype=XSD.float)
        elif isinstance(values["Value"], int):
            values["Value"] = Literal(values["Value"], datatype=XSD.integer)
        elif isinstance(values["Value"], bool):
            values["Value"] = Literal(values["Value"], datatype=XSD.boolean)
        elif isinstance(values["Value"], datetime):
            values["Value"] = Literal(values["Value"], datatype=XSD.dateTime)
        else:
            values["Value"] = Literal(values["Value"], datatype=XSD.string)

        return values

    @staticmethod
    def isfloat(x):
        try:
            _ = float(x)
        except (TypeError, ValueError):
            return False
        else:
            return True

    @staticmethod
    def isint(x):
        try:
            a = float(x)
            b = int(a)
        except (TypeError, ValueError):
            return False
        else:
            return a == b


class TransformationRules(RuleModel):
    metadata: Metadata
    classes: dict[str, Class]
    properties: dict[str, Property]
    prefixes: Optional[dict[str, Namespace]] = PREFIXES
    instances: Optional[list[Instance]] = None

    @property
    def raw_tables(self) -> list[str]:
        return list(
            {
                parse_rule(rule.rule, RuleType.rawlookup).table.name
                for rule in self.properties.values()
                if rule.is_raw_lookup
            }
        )

    @validator("properties", each_item=True)
    def class_property_exist(cls, value, values):
        if classes := values.get("classes"):
            if value.class_id not in classes:
                raise _exceptions.Error600(value.property_id, value.class_id).to_pydantic_custom_error()
        return value

    @validator("properties", each_item=True)
    def value_type_exist(cls, value, values):
        if classes := values.get("classes"):
            if value.property_type == "ObjectProperty" and value.expected_value_type not in classes:
                raise _exceptions.Error603(
                    value.class_i, value.property_id, value.expected_value_type
                ).to_pydantic_custom_error()
        return value

    @validator("properties")
    def is_type_defined_as_object(cls, value):
        defined_objects = {property_.class_id for property_ in value.values()}

        if undefined_objects := [
            property_.expected_value_type
            for _, property_ in value.items()
            if property_.property_type == "ObjectProperty" and property_.expected_value_type not in defined_objects
        ]:
            raise _exceptions.Error604(undefined_objects).to_pydantic_custom_error()
        else:
            return value

    @validator("prefixes")
    def are_prefixes_compliant(cls, value):
        if ill_formed_prefixes := [
            prefix for prefix, _ in value.items() if not re.match(prefix_compliance_regex, prefix)
        ]:
            raise _exceptions.Error400(ill_formed_prefixes, prefix_compliance_regex).to_pydantic_custom_error()
        else:
            return value

    @validator("prefixes")
    def are_namespaces_compliant(cls, value):
        ill_formed_namespaces = []
        for _, namespace in value.items():
            try:
                _ = URL(url=namespace).url
            except ValueError:
                ill_formed_namespaces += namespace

        if ill_formed_namespaces:
            raise _exceptions.Error401(ill_formed_namespaces).to_pydantic_custom_error()
        else:
            return value

    @validator("prefixes")
    def add_data_model_prefix_namespace(cls, value, values):
        if "metadata" not in values:
            raise _exceptions.Error601().to_pydantic_custom_error()
        if "prefix" not in values["metadata"].dict():
            raise _exceptions.Error602(missing_field="prefix").to_pydantic_custom_error()
        if "namespace" not in values["metadata"].dict():
            raise _exceptions.Error602(missing_field="namespace").to_pydantic_custom_error()

        value[values["metadata"].prefix] = values["metadata"].namespace
        return value



from cognite.neat.rules.models import TransformationRules


def get_instances_as_triples(transformation_rules: TransformationRules) -> list[tuple]:
    """Converts transformation rules instances sheet to RDF triples

    Parameters
    ----------
    transformation_rules : TransformationRules
        An instance of TransformationRules pydantic class

    Returns
    -------
    list[tuple]
        List of triples provided as tuples
    """
    if transformation_rules.instances:
        return [(instance.instance, instance.property_, instance.value) for instance in transformation_rules.instances]
    else:
        return []



from cognite.neat.rules import models
from cognite.neat.rules.parser import (
    parse_rules_from_excel_file,
    parse_rules_from_google_sheet,
    parse_rules_from_github_sheet,
    parse_rules_from_yaml,
)


__all__ = [
    "models",
    "parse_rules_from_excel_file",
    "parse_rules_from_google_sheet",
    "parse_rules_from_github_sheet",
    "parse_rules_from_yaml",
]
